# Add current project to sys path
import os
import sys
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, parent_dir)

from typing import List
import re
import logging
import shutil
import subprocess
import numpy as np

GEN_DIR = "pddl-generators"

def problem_blocksworld(seed: int = 123,
                        ops: int = 4,
                        num: int = 3) -> str:
    """
    See `util/pddl-generators/blocksworld/README`.
    :param seed: random seed
    :param ops: number of operators
    :param num: number of blocks
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/blocksworld/blocksworld {ops} {num} {seed}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    # Add `block` object type
    problem = "\n".join(
        [r if not r.startswith('(:objects') else r.strip()[:-1] + "- block)" for r in problem.split('\n')])

    # Rename `arm-empty` to `handempty`
    problem = problem.replace('arm-empty', 'handempty')

    # Rename `on-table` to `ontable`
    problem = problem.replace('on-table', 'ontable')

    # Rename domain from `blocksworld_4ops` to `blocksworld`
    problem = problem.replace('(:domain blocksworld-4ops)', '(:domain blocksworld)')

    return problem


def problem_depots(e: int = 1,
                   i: int = 1,
                   t: int = 2,
                   p: int = 2,
                   h: int = 2,
                   c: int = 2,
                   seed: int = 123) -> str:
    """
    See `util/pddl-generators/depots/README.txt`.
    :param e: number of depots
    :param i: number of distributors
    :param t: number of trucks
    :param p: number of pallets
    :param h: number of hoists
    :param c: number of crates
    :param seed: random seed
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/depots/depots -e {e} -i {i} -t {t} -p {p} -h {h} -c {c} -s {seed}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    return problem


def problem_driverlog(seed: int = 123,
                      road_junctions: int = 1,
                      drivers: int = 1,
                      packages: int = 1,
                      trucks: int = 1) -> str:
    """
    See `util/pddl-generators/driverlog/README.txt`.
    :param seed: random seed
    :param road_junctions: total number of connections between all location pairs
    :param drivers: number of drivers
    :param packages: number of packages
    :param trucks: number of trucks
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/driverlog/dlgen -t {seed} {road_junctions} {drivers} {packages} {trucks}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    # Remove cost metric
    problem = problem.replace("(:metric minimize (total-time))", "")

    return problem


def problem_barman(seed: int = 123,
                   num_cocktails: int = 2,
                   num_ingredients: int = 3,
                   num_shots: int = 3) -> str:
    """
    See `util/pddl-generators/barman/README.txt`.
    :param seed: random seed
    :param num_cocktails: number of cocktails
    :param num_ingredients: number of ingredients
    :param num_shots: number of shots
    :return: problem string
    """
    # Generate a problem
    result = subprocess.run(f"python {GEN_DIR}/barman/barman-generator.py {num_cocktails} {num_ingredients} {num_shots} {seed}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    return problem


def problem_visitall(seed: int = 123,
                     x: int = 3,
                     y: int = 3,
                     r: int = 0.2,
                     u: int = 2) -> str:
    """
    See `util/pddl-generators/visitall/README.txt`.
    :param seed: random seed
    :param x: number of grid cols
    :param y: number of grid rows
    :param r: ratio of cells to be visited
    :param u: number of unavailable grid cells
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/visitall/grid -x {x} -y {y} -r {r} -u {u} -s {seed}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    problem = problem[problem.find('(define '):]

    return problem


def problem_ferry(seed: int = 123,
                  l: int = 2,
                  c: int = 1) -> str:
    """
    See `util/pddl-generators/ferry/README.txt`.
    :param seed: random seed
    :param l: number of locations
    :param c: number of cars
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/ferry/ferry -l {l} -c {c} -s {seed}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    # Add location and car types
    refactored_problem = []
    rows = problem.split('\n')
    for i in range(len(rows)):
        r = rows[i]
        if 'objects' in r:
            refactored_problem.append(f"{r} - location")
        elif 'objects' in rows[i-1]:
            refactored_problem.append(f"{r} - car")
        else:
            if not r.startswith('(location ') and not r.startswith('(car '):
                refactored_problem.append(r)
    problem = "\n".join(refactored_problem)

    # Ferry: hyphens '-' are not yet well supported in unified-planning
    # --> change not-eq into not_eq --> remove not_ for OffLAM
    # --> TODO 1: refactor OffLAM
    # --> TODO 2: open issue in unified-planning
    problem = problem.replace('(not-eq ', '(noteq ')

    return problem


def problem_floortile(seed: int = 123,
                      num_rows: int = 2,
                      num_columns: int = 2,
                      num_robots: int = 2,
                      mode_flag: str = 'seq') -> str:
    """
    See `util/pddl-generators/floortile/README.txt`.
    :param seed: random seed
    :param num_rows: number of rows
    :param num_columns: number of columns
    :param num_robots: number of robots
    :param mode_flag: either `seq` or `time`
    :return: problem string
    """

    # Generate a problem
    problem_name = f"floortile-r{num_rows}-c{num_columns}-robot{num_robots}-s{seed}"
    result = subprocess.run(f"./{GEN_DIR}/floortile/floortile-generator.py {problem_name} {num_rows} {num_columns} {num_robots} {mode_flag} {seed}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    # Remove half goals to make the problem almost surely feasible
    goal_block = re.search(r'\(:goal\s*\(and(.*?)\)\s*\)', problem, re.DOTALL)
    if not goal_block:
        raise Exception

    # get goal facts
    goals = re.findall(r'\([^)]+\)', goal_block.group(1))

    # extract a random subset of goal facts
    subgoals = np.random.choice(goals, int(len(goals)/2))

    # Reconstruct the goal block
    new_goal_block = '(:goal (and\n    ' + '\n    '.join(subgoals) + '\n)'

    # Replace original goal block in text
    problem = re.sub(r'\(:goal\s*\(and\s*((?:.|\n)*?)\s*\)\s*\)', new_goal_block, problem, re.DOTALL)

    # Remove total cost a cost minimization
    problem = problem.replace('(= (total-cost) 0)', '')
    problem = problem.replace('(:metric minimize (total-cost))', '')

    return problem


def problem_goldminer(seed: int = 123,
                      r: int = 2,
                      c: int = 2) -> str:
    """
    See `util/pddl-generators/goldminer/README`.
    :param seed: random seed
    :param r: number of rows
    :param c: number of columns
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/goldminer/gold-miner-generator -r {r} -c {c} -s {seed}".split(),
                            capture_output=True, text=True)
    problem = f"{result.stdout}\n)"

    # Add (and ...) to goal definition (which involves a single literal)
    problem = re.sub(r'(:goal\s*)\(([^()]+)\)', r'\1(and (\2))', problem)

    return problem


def problem_grid(seed: int = 123,
                 x: int = 1,
                 y: int = 1,
                 t: int = 1,
                 p: List[int] = np.array([100]),
                 k: List[int] = np.array([1]),
                 l: List[int] = np.array([1])) -> str:
    """
    See `util/pddl-generators/grid/README`.
    :param seed: random seed
    :param x: horizontal extension of grid
    :param y: vertical extension of grid
    :param t: number of different key and lock types
    :param p: probability, for any key, to be mentioned in the goal
    :param k: number of keys vector (one 0 ... 9 entry for each type)
    :param l: number of locks vector (one 0 ... 9 entry for each type)
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/grid/grid  -x {x} -y {y} -t {t} "
                            f"-k {str(k)[1:-1].replace(',', '').replace(' ', '')} "
                            f"-l {str(l)[1:-1].replace(',', '').replace(' ', '')} "
                            f"-s {seed}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    # Add object types
    refactored_problem = []
    rows = problem.split('\n')
    for i in range(len(rows)):
        if ' f0-' in rows[i] and rows[i+1].strip().startswith('shape0'):
            refactored_problem.append(f"{rows[i]} - place")
        elif ' shape' in rows[i] and rows[i+1].strip().startswith('key0-'):
            refactored_problem.append(f"{rows[i]} - shape")
        elif rows[i].strip().startswith('key') and rows[i+1] == ')':
            refactored_problem.append(f"{rows[i]} - key")
        elif '(key ' not in rows[i] and "(place " not in rows[i] and "(shape " not in rows[i]:
            refactored_problem.append(rows[i])
    problem = '\n'.join(refactored_problem)
    return problem


def problem_grippers(seed: int = 123,
                     n: int = 1,
                     r: int = 1,
                     o: int = 1) -> str:
    """
    See `util/pddl-generators/grippers/README.txt`.
    :param seed: random seed
    :param n: number of robots
    :param r: number of rooms
    :param o: number of balls
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/grippers/grippers  -n {n} -r {r} -o {o} -s {seed}".split(),
                            capture_output=True, text=True)
    problem = result.stdout
    problem = problem.replace(' object', ' ball')  # unified-planning does not support type `object`
    return problem


def problem_hanoi(seed: int = 123,
                  n: int = 1) -> str:
    """
    See `util/pddl-generators/hanoi/README.txt`.
    :param seed: random seed (not used since the problems generator is not randomized)
    :param n: number of discs
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/hanoi/hanoi  -n {n}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    # Add object types
    refactored_problem = []
    rows = problem.split('\n')
    for i in range(len(rows)):
        if rows[i].startswith('(:objects'):
            platforms = [w for w in rows[i].split()[1:-1] if w.startswith('peg')]
            discs = [w for w in rows[i].split()[1:-1] if w.startswith('d')]
            refactored_problem.append(f"(:objects \n{' '.join(platforms)} - platform\n {' '.join(discs)} - disc\n)")
        else:
            refactored_problem.append(rows[i])
    problem = '\n'.join(refactored_problem)
    return problem


def problem_matchingbw(seed: int = 123,
                       n: int = 1) -> str:
    """
    See `util/pddl-generators/hanoi/README.txt`.
    :param seed: random seed
    :param n: number of blocks
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/matchingbw/matching-bw-generator.sh tmp {n} {seed}".split(),
                            capture_output=True, text=True)
    with open(f"{GEN_DIR}/matchingbw/tmp-typed.pddl", 'r') as f:
        problem = f.read()
    os.remove(f"{GEN_DIR}/matchingbw/tmp-typed.pddl")
    os.remove(f"{GEN_DIR}/matchingbw/tmp-untyped.pddl")

    return problem


def problem_miconic(seed: int = 123,
                       f: int = 1,
                       p: int = 1) -> str:
    """
    See `util/pddl-generators/miconic/README.txt`.
    :param seed: random seed
    :param f: number of floors
    :param p: number of passengers
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/miconic/miconic -f {f} -p {p} -r {seed}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    return problem


def problem_npuzzle(seed: int = 123,
                    n: int = 1) -> str:
    """
    See `util/pddl-generators/npuzzle/README.txt`.
    :param seed: random seed
    :param n: number of rows/columns
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/npuzzle/n-puzzle-generator -n {n} -s {seed}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    return problem


def problem_nomystery(seed: int = 123,
                      l: int = 1,
                      p: int = 1,
                      n: int = 1,
                      m: int = 1,
                      c: int = 1) -> str:
    """
    See `util/pddl-generators/nomystery/README.txt`.
    :param seed: random seed
    :param l: number of locations
    :param p: number of packages
    :param n: edges ratio in location graph (i.e. total edges = n * l)
    :param m: maximum edges weight
    :param c: constrainedness (initial fuel supply = c * optimal_cost)
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/nomystery/nomystery -l {l} -p {p} -n {n} -m {m} -c {c} -s {seed} -e 0".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    # Remove total cost a cost minimization
    problem = problem.replace('(= (total-cost) 0)', '')
    problem = problem.replace('(:metric minimize (total-cost))', '')

    return problem


def problem_parking(seed: int = 123,
                    curbs: int = 1,
                    cars: int = 1) -> str:
    """
    See `util/pddl-generators/parking/README.txt`.
    :param seed: random seed
    :param curbs: number of curbs
    :param cars: number of cars
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"perl ./{GEN_DIR}/parking/parking-generator.pl _ {curbs} {cars} seq {seed}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    # Remove PDDL comments
    problem = '\n'.join([r for r in problem.split('\n') if not r.strip().startswith(';')])

    return problem


def problem_rovers(seed: int = 123,
                   r: int = 1,
                   w: int = 1,
                   o: int = 1,
                   c: int = 1,
                   g: int = 1) -> str:
    """
    See `util/pddl-generators/rovers/README`.
    :param seed: random seed
    :param r: number of rovers
    :param w: number of waypoints
    :param o: number of objectives
    :param c: number of cameras
    :param g: number of goals
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/rovers/rovgen {seed} {r} {w} {o} {c} {g}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    return problem


def problem_satellite(seed: int = 123,
                      s: int = 1,
                      i: int = 1,
                      m: int = 1,
                      t: int = 1,
                      o: int = 1) -> str:
    """
    See `util/pddl-generators/satellite/README`.
    :param seed: random seed
    :param s: number of satellites
    :param i: maximum number of instruments
    :param m: number of modes
    :param t: number of targets
    :param o: number of observations
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/satellite/satgen {seed} {s} {i} {m} {t} {o}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    return problem


def problem_sokoban(seed: int = 123,
                    n: int = 1,
                    b: int = 1,
                    w: int = 1) -> str:
    """
    See `util/pddl-generators/sokoban/README.txt`.
    :param seed: random seed
    :param n: grid size
    :param b: number of boxes
    :param w: number of walls
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/sokoban/random/sokoban-generator-typed -n {n} -b {b} -w {w} -s {seed}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    # Remove comments
    problem = '\n'.join([r for r in problem.split('\n') if not r.startswith(';')])

    # It may happen that the
    while len([r for r in problem.split('\n') if r.strip() != '']) == 0:
        logging.debug('An error occured during sokoban problem generation. Retrying to generate a new problem.')
        seed += 1
        result = subprocess.run(f"./{GEN_DIR}/sokoban/random/sokoban-generator-typed -n {n} -b {b} -w {w} -s {seed}".split(),
                                capture_output=True, text=True)
        problem = result.stdout
        # Remove comments
        problem = '\n'.join([r for r in problem.split('\n') if not r.startswith(';')])

    return problem


def problem_spanner(seed: int = 123,
                    spanners: int = 1,
                    nuts: int = 1,
                    locations: int = 3) -> str:
    """
    See `util/pddl-generators/spanner/README.txt`.
    :param seed: random seed
    :param spanners: number of spanners
    :param nuts: number of nuts
    :param locations: number of locations
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"python ./{GEN_DIR}/spanner/spanner-generator.py --seed {seed}"
                            f" {spanners} {nuts} {locations}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    return problem


def problem_tpp(seed: int = 123,
                p: int = 1,
                m: int = 1,
                t: int = 3,
                d: int = 3,
                l: int = 3) -> str:
    """
    See `util/pddl-generators/tpp/README`.
    :param seed: random seed
    :param p: number of products
    :param m: number of markets
    :param t: number of trucks
    :param d: number of depots
    :param l: maximum goods level
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/tpp/tpp -s {seed} -p {p} -m {m} -t {t} -d {d} -l {l} tmp.pddl".split(),
                            capture_output=True, text=True)

    with open(f"tmp.pddl", 'r') as f:
        problem = f.read()
    os.remove(f"tmp.pddl")

    return problem


def problem_transport(seed: int = 123,
                      generator: str = 'city',
                      n: int = 1,
                      size: int = 10,
                      degree: int = 3,
                      mindistance: int = 10,
                      trucks: int = 3,
                      packages: int = 3) -> str:
    """

    See `util/pddl-generators/transport/README.txt`.
    :param seed: random seed
    :param generator: transport generator type in [city, two-cities, three-cities]
    :param n: number of nodes
    :param size: size for computing `connect_distance` = math.sqrt((degree * size * size) / (nodes * math.pi * 0.694))
    :param degree: degree for computing `connect_distance` = math.sqrt((degree * size * size) / (nodes * math.pi * 0.694))
    :param mindistance: minimum distance between two nodes
    :param trucks: number of trucks
    :param packages: number of packages
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"python ./{GEN_DIR}/transport/{generator}-generator.py"
                            f" {n} {size} {degree} {mindistance} {trucks} {packages} {seed}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    # Remove total cost a cost minimization
    problem = problem.replace('(= (total-cost) 0)', '')
    problem = problem.replace('(:metric minimize (total-cost))', '')

    # Remove road lengths and comments
    problem = '\n'.join([r for r in problem.split('\n')
                         if not r.strip().startswith('(= ') and not r.strip().startswith(';')])

    return problem


def problem_zenotravel(seed: int = 123,
                       cities: int = 1,
                       planes: int = 1,
                       people: int = 3,
                       distance: int = 1) -> str:
    """
    See `util/pddl-generators/zenotravel/README`.
    :param seed: random seed
    :param cities: number of cities
    :param planes: number of planes
    :param people: number of people
    :param distance: numerical distance between cities
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"./{GEN_DIR}/zenotravel/ztravel {seed} {cities} {planes} {people}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    return problem


def problem_childsnack(seed: int = 123,
                       children: int = 1,
                       trays: int = 1,
                       gluten_factor: int = 3,
                       const_ratio: int = 1.3) -> str:
    """
    See `util/pddl-generators/childsnack/README.txt`.
    :param seed: random seed
    :param children: number of children
    :param trays: number of trays
    :param gluten_factor: gluten ratio among children
    :param const_ratio: proportion of needed symbols that are declared in advance in the problem file.
    The min ratio should be 1.0 to guarantee solvability.
    :return: problem string
    """

    # Generate a problem
    result = subprocess.run(f"python ./{GEN_DIR}/childsnack/child-snack-generator.py pool "
                            f"{seed} {children} {trays} {gluten_factor} {const_ratio}".split(),
                            capture_output=True, text=True)
    problem = result.stdout

    return problem


def problem_elevators(seed: int = 123,
                      floors: int = 2,
                      area_size: int = 2,
                      fast_elevators: int = 1,
                      slow_elevators: int = 1,
                      fast_capacity: int = 1,
                      slow_capacity: int = 3,
                      passengers: int = 2) -> str:
    """
    See `util/pddl-generators/elevators/README.txt`.
    :param seed: random seed
    :param floors: number of floors
    :param area_size: area size (must be a factor of floors, otherwise passengers can start in out of bounds floors)
    :param fast_elevators: number of fast elevators
    :param slow_elevators: number of slow elevators
    :param fast_capacity: passengers capacity of every fast elevators
    :param slow_capacity: passengers capacity of every slow elevators
    :param passengers: number of passengers
    :return: problem string
    """

    def generate_executable(c_file_path: str, **kwargs):
        # Path to your C source file
        # c_file ==
        # executable = "./elevator_problem"  # Linux/macOS, use "elevator_problem.exe" for Windows

        c_file = 'my_generate_data.c'
        my_c_file = f"{os.path.join(*c_file_path.split('/')[:-1])}/{c_file}"
        shutil.copy(c_file_path, my_c_file)

        # Constants that can be changed
        constants = {"FLOORS", "AREA_SIZE", "FAST_ELEVATORS", "SLOW_ELEVATORS", "FAST_CAPACITY", "SLOW_CAPACITY"}
        assert set(kwargs.keys()).issubset({c.lower() for c in constants})

        # modify constants in the C file
        with open(my_c_file, "r") as f:
            c_code = f.read()
        for const, new_value in kwargs.items():
            pattern = rf"#define\s+{const.upper()}\s+\d+"
            replacement = f"#define {const.upper()} {new_value}"
            c_code = re.sub(pattern, replacement, c_code)

        # make generation reproducible by allowing for a random seed as input argument
        c_code = c_code.replace('MinPassengers=atoi(argv[1]);', 'srand((unsigned int)atoi(argv[1]));\nMinPassengers=atoi(argv[2]);')
        c_code = c_code.replace('MaxPassengers=atoi(argv[2]);', 'MaxPassengers=atoi(argv[3]);')
        c_code = c_code.replace('Step=atoi(argv[3]);', 'Step=atoi(argv[4]);')
        c_code = c_code.replace('MinID=atoi(argv[4]);', 'MinID=atoi(argv[5]);')
        c_code = c_code.replace(f'MaxID=atoi(argv[5]);', f'MaxID=atoi(argv[6]);')
        c_code = c_code.replace("if (argc!=6) {", "if (argc!=7) {")
        c_code = c_code.replace('printf("MinPassengers, MaxPassenegrs, Step, MinID, MaxID\n\nwhere:\n\n");',
                       'printf("Seed, MinPassengers, MaxPassenegrs, Step, MinID, MaxID\n\nwhere:\n\n");'
                       '\nprintf("Seed : random seed\n");')
        c_code = c_code.replace("srand( (unsigned)time( NULL ) );", "")

        # save the updated C file
        with open(my_c_file, "w") as file:
            file.write(c_code)

        # compile the C file
        executable_file_path = my_c_file.replace(".c", "")
        compile_cmd = ["gcc", my_c_file, "-o", executable_file_path]
        result = subprocess.run(compile_cmd, capture_output=True, text=True)

        if result.returncode != 0:
            print("Compilation failed:\n", result.stderr)
            exit(1)
        return executable_file_path

    # Generate a problem
    executable_path = generate_executable(f"./{GEN_DIR}/elevators/generate_data.c",
        **dict(floors=floors, area_size=area_size, fast_elevators=fast_elevators, slow_elevators=slow_elevators,
             fast_capacity=fast_capacity, slow_capacity=slow_capacity))
    # Generate problem in txt
    result = subprocess.run(f"{executable_path} {seed} {passengers} {passengers} 1 1 1".split(),
                            capture_output=True, text=True)
    # Parse problem in pddl
    result = subprocess.run(f"./{GEN_DIR}/elevators/generate_pddl {floors} {floors} 1 {passengers} {passengers} 1 1 1".split(),
                            capture_output=True, text=True)

    prob_file = [f for f in os.listdir(f"./") if f.endswith('.pddl') and f.startswith('p')][0]
    with open(prob_file, 'r') as f:
        problem = f.read()
    os.remove(prob_file)
    os.remove(prob_file.replace('.pddl', '.txt'))

    # Remove cost functions
    problem = '\n'.join([r for r in problem.split('\n')
                         if ':metric' not in r and not r.startswith('(=')])

    return problem



#!/usr/bin/env python

import os.path
import subprocess
import sys

JAR_NAME = "tidybot.jar"

worldsize = sys.argv[1]
ntables = sys.argv[2]
ncupboards = sys.argv[3]
minsurfacesize = sys.argv[4]
maxsurfacesize = sys.argv[5]
cupboardsize = sys.argv[6]

script_dir = os.path.abspath(os.path.dirname(__file__))

cmd = [
    "java",
    "-jar",
    f"{script_dir}/{JAR_NAME}",
    worldsize,
    ntables,
    ncupboards,
    minsurfacesize,
    maxsurfacesize,
    cupboardsize,
]

if len(sys.argv) == 8:
    cmd.append(sys.argv[7])

subprocess.check_call(cmd)



import difflib
import inspect
from pathlib import Path
import pkgutil
import importlib

registry = {}

package_dir = Path(__file__).resolve().parent
for _, module_name, is_pkg in pkgutil.iter_modules([str(package_dir)]):

    if is_pkg:
        continue
    module = importlib.import_module(f"{__name__}.{module_name}")

    # Expose algorithm classes to allow import such as: from amlgym.algorithms import OffLAM
    if 'AlgorithmAdapter' not in module_name:
        class_obj = getattr(module, module_name, None)
        assert class_obj is not None, f"{module_name}.{module_name} class is not defined"

        for name, obj in inspect.getmembers(module, inspect.isclass):
            if name.lower() == module_name.lower():
                registry[name.lower()] = obj


def get_algorithm(name, **kwargs):
    """
    Retrieve an algorithm by name from the registry.

    If the name is not found, raises a ValueError with suggestions for close matches.
    """
    try:
        return registry[name.lower()](**kwargs)
    except KeyError:
        # Find close matches to the requested name
        matches = difflib.get_close_matches(name, registry.keys(), n=3, cutoff=0.6)
        suggestions = f" Did you mean: {', '.join(matches)}?" if matches else ""
        available = ", ".join(sorted(registry.keys()))
        raise ValueError(
            f"Unknown algorithm '{name}'. Available algorithms: {available}.{suggestions}"
        ) from None


def print_algorithms() -> None:
    """
    Print available algorithms and their constructor parameters.
    """
    print("================== AMLGym algorithms suite ==================")
    for i, (name, cls) in enumerate(sorted(registry.items())):
        sig = inspect.signature(cls.__init__)
        params = [
            str(p)
            for pname, p in sig.parameters.items()
            if pname != "self"
            and str(p) != "**kwargs"
        ]

        # Print algorithm name
        print(f"\n{i + 1} - {name}({', '.join(params)})")

        # Print reference
        ref = getattr(cls, "_reference", None)

        if isinstance(ref, dict):
            ref_str = "\n".join(f"\x1B[3m\033[1m\t{k}: {v}.\033[0m" for k, v in ref.items())
        else:
            ref_str = "\tNo reference provided"

        print(ref_str)



#! /usr/bin/env python
#
# Usage:
# $./generator.py [number of data items] [number of layers] [number of scripts] network
#
#
# File: network
# --------------------------------------------------------------------
# 1 2 4 8 16 32 % file sizes
# 0.1 0.1 0.2 0.3 0.2 0.1 % probability distribution or frequency distribution over file sizes (write 0 for uniform distribution)
#
# % list of servers
# 56 % capacity
# 1 % base io cost
# 1 2  % base processing costs
# 0.2 0.2 0.6 % probability distribution over processing costs with additional probability for disabled script (write 0 for uniform distribution)
#
# 32
# 2
# 2
# 1.0 0.0
#
# 96
# 1
# 1 3
# 0.1 0.9 0.0
#
# 96
# 2
# 2
# 0.5 0.5
#
# % list of bidirectional connections
# 1 2  % adjacent servers
# 1 % base send cost
#
# 2 3
# 2
#
# 1 3
# 3
#
# 1 4
# 2
# -------------------------------------------------------------------

import sys
import numpy
import re
import importlib
from collections import OrderedDict


class DistributedComputing:
    def __init__(self, num_data_items, num_layers, num_scripts, filename, seed):
        self.layer_distr = []
        self.servers = {}
        self.connections = {}
        self.initial_data_items = []
        self.goal_data_items = []
        self.task_name = ""
        self.max_capacity = 0
        self.data_items = {}
        self.scripts = {}
        self.server_script_pairs = {}
        self.server_data_item_pairs_init = []
        self.server_data_item_pairs_goal = []
        self.process_costs = {}
        self.send_costs = {}
        self.io_costs = {}

        self.num_data_items = num_data_items
        self.num_layers = num_layers
        self.num_scripts = num_scripts
        self.seed = seed

        numpy.random.seed(seed)

        self.task_name = f"prob-{num_data_items}-{num_layers}-{num_scripts}-{filename}-{seed}"
        self._parse_module(filename)

        self._check_input_values()
        self._generate_data_items()
        self._assign_sizes_to_data_items()
        self._generate_scripts()
        self._determine_max_capacity()
        self._check_required_capacity()
        self._assign_scripts_to_servers()
        self._assign_initial_data_items_to_servers()
        self._assign_goal_data_items_to_servers()
        self._compute_process_costs()
        self._compute_send_costs()
        self._compute_io_costs()
        self._write_pddl()

    def _check_input_values(self):
        if self.num_data_items <= self.num_layers:
            print("[number of layers] must be smaller than [number of data items]")
            sys.exit(0)
        if self.num_scripts < self.num_data_items - 2:
            print(
                "[number of scripts] must be larger or equal than [number of data items]-2"
            )
            sys.exit(0)

    def _parse_module(self, module):
        m = importlib.import_module(module)

        self.data_size_distr = getattr(m, "FILE_SIZE_DISTRIBUTION")
        self.servers = getattr(m, "SERVERS")
        self.connections = getattr(m, "CONNECTIONS")

    def _clean_line(self, line):
        line = line.split("%", 1)[0]
        line = line.rstrip()
        return line

    def _skip_empty_lines(self, lines, line_nr):
        while True:
            try:
                line = self._clean_line(lines[line_nr])
                if line != "":
                    return line_nr
                line_nr += 1
            except IndexError:
                raise
        return line_nr

    def _read_numbers(self, line, line_nr, error_attribute):
        numbers = []
        try:
            for size in line.split(" "):
                numbers.append(int(size))
            if not numbers:
                raise ValueError
        except ValueError:
            print(
                "Expected '{}' of line {} to be a space separated list of {}.".format(
                    line, line_nr, error_attribute
                )
            )
            sys.exit(0)
        return numbers

    def _read_distribution(self, line, line_nr):
        probabilities = []
        try:
            for size in line.split(" "):
                probabilities.append(float(size))
            if not probabilities:
                raise ValueError
        except ValueError:
            print(
                "Expected '{}' of line {} to be 0 or a space separated list of probabilities or frequencies.".format(
                    line, line_nr
                )
            )
            sys.exit(0)
        return probabilities

    def _generate_data_items(self):
        for i in range(self.num_data_items):
            self.data_items[f"data{i + 1}"] = {}

    def _assign_sizes_to_data_items(self):
        self.data_sizes = set()
        for data_item in self.data_items:
            size = self.data_size_distr()
            self.data_items[data_item]["size"] = size
            self.data_sizes.add(size)
        self.data_sizes = list(self.data_sizes)

    def _generate_scripts(self):
        layers = self._distribute_data_items_to_layers()
        self.initial_data_items = layers[0][:]
        items_used_in_scripts = set()
        script_nr = 1
        # Generate at one script for each data item
        for layer_nr in range(1, self.num_layers):
            for data_item in layers[layer_nr]:
                script = f"script{script_nr}"
                properties = {}
                input1 = numpy.random.choice(layers[layer_nr - 1])
                # Enforce different data inputs
                input2 = input1
                while input2 == input1:
                    random_layer = numpy.random.choice(layer_nr)
                    input2 = numpy.random.choice(layers[random_layer])
                items_used_in_scripts.add(input1)
                items_used_in_scripts.add(input2)

                properties["input1"] = input1
                properties["input2"] = input2
                properties["output"] = data_item
                self.scripts[script] = properties
                script_nr += 1
        # Distribute remaining scripts among layers
        while script_nr <= self.num_scripts:
            layer_nr = numpy.random.choice(range(1, self.num_layers))
            output = numpy.random.choice(layers[layer_nr])
            script = f"script{script_nr}"
            properties = {}
            input1 = numpy.random.choice(layers[layer_nr - 1])
            # Enforce different data inputs
            input2 = input1
            while input2 == input1:
                random_layer = numpy.random.choice(layer_nr)
                input2 = numpy.random.choice(layers[random_layer])
            items_used_in_scripts.add(input1)
            items_used_in_scripts.add(input2)
            properties["input1"] = input1
            properties["input2"] = input2
            properties["output"] = output

            self.scripts[script] = properties
            script_nr += 1

        self.goal_data_items = [
            item
            for layer in layers
            for item in layer
            if not item in items_used_in_scripts
        ]

    def _distribute_data_items_to_layers(self):
        distributed_items = 0
        layers = []
        for data_item in self.data_items:
            # Enforce at least one data item on each layer
            if distributed_items < self.num_layers:
                layers.append([])
                layers[distributed_items].append(data_item)
            # First layer needs at least two items
            elif distributed_items == self.num_layers:
                layers[0].append(data_item)
            else:
                # TODO: Use other distribution than uniform distribution
                layer = numpy.random.choice(self.num_layers)
                layers[layer].append(data_item)
            distributed_items += 1
        layers = self._add_layer_to_data_item_names(layers)
        return layers

    def _add_layer_to_data_item_names(self, layers):
        renamed_data_items = {}
        renamed_on_layers = []
        layer_nr = 0
        for layer in layers:
            renamed_on_layers.append([])
            for data_item in layer:
                renamed_data_item = "data-{}-{}".format(
                    layer_nr, data_item.replace("data", "")
                )
                renamed_data_items[renamed_data_item] = self.data_items[data_item]
                renamed_on_layers[layer_nr].append(renamed_data_item)
            layer_nr += 1
        self.data_items = renamed_data_items
        return renamed_on_layers

    def _determine_max_capacity(self):
        self.max_capacity = max(
            properties["capacity"] for server, properties in self.servers.items()
        )

    def _check_required_capacity(self):
        min_space = max(value["size"] for (key, value) in self.data_items.items())
        max_space = max(
            
                self.data_items[value["input1"]]["size"]
                + self.data_items[value["input2"]]["size"]
                + self.data_items[value["output"]]["size"]
                for (key, value) in self.scripts.items()
            
        )
        for server, properties in self.servers.items():
            capacity = properties["capacity"]
            if capacity < min_space:
                print(
                    "Server {} has not enough capacity to cache data item of size {}.".format(
                        server, min_space
                    )
                )
                sys.exit(0)
        if max_space > self.max_capacity:
            print(
                "Any server has not enough capacity to process script that requires {} space.".format(
                    max_space
                )
            )
            sys.exit(0)

    def _assign_scripts_to_servers(self):
        try:
            for server, properties in self.servers.items():
                for script, value in self.scripts.items():
                    self.server_script_pairs[(server, script)] = {
                        "base_processing_cost": properties["processing_distr"]()
                    }
        except ValueError:
            print("Expect 0 or probabilities sum to 1.")
            sys.exit(0)

    def _assign_initial_data_items_to_servers(self):
        servers = self.servers.keys()
        for data_item in self.initial_data_items:

            server = numpy.random.choice([s for s in servers])
            self.server_data_item_pairs_init.append((server, data_item))

    def _assign_goal_data_items_to_servers(self):
        servers = self.servers.keys()
        for data_item in self.goal_data_items:
            server = numpy.random.choice([s for s in servers])

            self.server_data_item_pairs_goal.append((server, data_item))

    def _compute_process_costs(self):
        for pair, props in self.server_script_pairs.items():
            if props["base_processing_cost"] >= 0:
                self.process_costs[(pair[1], pair[0])] = props["base_processing_cost"]

    def _compute_send_costs(self):
        for data_size in self.data_sizes:
            for connection, props in self.connections.items():
                self.send_costs[
                    (
                        (
                            f"server{connection[0]}",
                            f"server{connection[1]}",
                        ),
                        f"number{data_size}",
                    )
                ] = (data_size * props["send"])

    def _compute_io_costs(self):
        for data_size in self.data_sizes:
            for server, props in self.servers.items():
                self.io_costs[(server, f"number{data_size}")] = (
                    data_size * props["io"]
                )

    def _write_pddl(self):
        print(
            "(define (problem p{}-{}-{}-{}-{})".format(
                self.num_data_items,
                self.num_layers,
                self.num_scripts,
                self.task_name,
                self.seed,
            )
        )
        print("    (:domain data-network)")
        self._write_objects()
        self._write_initial_state()
        self._write_goals()
        print("    (:metric minimize (total-cost))")
        print(")")

    def _write_objects(self):
        data_items = sorted(
            (data_item for data_item in self.data_items),
            key=lambda x: self._natural_keys(x),
        )
        print("    (:objects")
        for data_item in data_items[:-1]:
            print(f"              {data_item}")
        print(f"              {data_items[-1]} - data")

        scripts = sorted(
            (script for script in self.scripts), key=lambda x: self._natural_keys(x)
        )
        for script in scripts[:-1]:
            print(f"              {script}")
        print(f"              {scripts[-1]} - script")
        servers = sorted(
            (server for server in self.servers), key=lambda x: self._natural_keys(x)
        )
        for server in servers[:-1]:
            print(f"              {server}")
        print(f"              {servers[-1]} - server")
        for number in range(0, self.max_capacity):
            print(f"              number{number}")
        print(f"              number{self.max_capacity} - numbers")
        print(f"    )")

    def _write_initial_state(self):
        print("    (:init")
        self._write_constant_predicates()
        self._write_cost_predicates()
        self._write_predicates()
        print("    )")

    def _write_constant_predicates(self):
        scripts = OrderedDict(
            sorted(self.scripts.items(), key=lambda x: self._natural_keys(x[0]))
        )
        for script, properties in scripts.items():
            print(
                "           (SCRIPT-IO {} {} {} {})".format(
                    script,
                    properties["input1"],
                    properties["input2"],
                    properties["output"],
                )
            )
        script_server_pairs = sorted(
            (
                (pair[1], pair[0])
                for pair in self.server_script_pairs
                if self.server_script_pairs[pair]["base_processing_cost"] >= 0
            ),
            key=lambda x: self._natural_keys(x[0] + x[1]),
        )
        connections = sorted(connection for connection in self.connections)
        for connection in connections:
            print(
                "           (CONNECTED server{} server{})".format(
                    connection[0], connection[1]
                )
            )
            print(
                "           (CONNECTED server{} server{})".format(
                    connection[1], connection[0]
                )
            )
        data_items = OrderedDict(
            sorted(self.data_items.items(), key=lambda x: self._natural_keys(x[0]))
        )
        for data_item, properties in data_items.items():
            print(
                "           (DATA-SIZE {} number{})".format(
                    data_item, properties["size"]
                )
            )
        servers = OrderedDict(
            sorted(self.servers.items(), key=lambda x: self._natural_keys(x[0]))
        )
        for server, properties in servers.items():
            print(
                "           (CAPACITY {} number{})".format(
                    server, properties["capacity"]
                )
            )
        self._write_math()

    def _write_math(self):
        for i in range(0, self.max_capacity + 1):
            for j in sorted(self.data_sizes):
                if i + j <= self.max_capacity:
                    print(
                        "           (SUM number{} number{} number{})".format(
                            i, j, i + j
                        )
                    )
        capacities = sorted(
            {value["capacity"] for (key, value) in self.servers.items()}
        )
        for i in range(1, self.max_capacity + 1):
            for j in capacities:
                if i <= j:
                    print(
                        "           (LESS-EQUAL number{} number{})".format(
                            i,
                            j,
                        )
                    )

    def _write_predicates(self):
        init_server_data = sorted(
            ((pair[1], pair[0]) for pair in self.server_data_item_pairs_init),
            key=lambda x: self._natural_keys(x[0]),
        )
        for pair in init_server_data:
            print(f"           (saved {pair[0]} {pair[1]})")
        servers = sorted(
            (server for server in self.servers), key=lambda x: self._natural_keys(x)
        )
        for server in servers:
            print(f"           (usage {server} number0)")

    def _write_cost_predicates(self):
        print("           (= (total-cost) 0)")
        process_costs = OrderedDict(
            sorted(
                self.process_costs.items(),
                key=lambda x: self._natural_keys(f"{x[0]}{x[1]}"),
            )
        )
        for pair, value in process_costs.items():
            print(
                f"           (= (process-cost {pair[0]} {pair[1]}) {value})"
            )
        send_costs = OrderedDict(
            sorted(
                self.send_costs.items(),
                key=lambda x: self._natural_keys(
                    f"{x[0][0]}{x[0][1]}{x[1]}"
                ),
            )
        )
        for pair, value in send_costs.items():
            print(
                "           (= (send-cost {} {} {}) {})".format(
                    pair[0][0], pair[0][1], pair[1], value
                )
            )
            print(
                "           (= (send-cost {} {} {}) {})".format(
                    pair[0][1], pair[0][0], pair[1], value
                )
            )
        io_costs = OrderedDict(
            sorted(
                self.io_costs.items(),
                key=lambda x: self._natural_keys(f"{x[0]}{x[1]}"),
            )
        )
        for pair, value in io_costs.items():
            print(f"           (= (io-cost {pair[0]} {pair[1]}) {value})")

    def _write_goals(self):
        print("    (:goal")
        print("           (and")
        goal_server_data = sorted(
            ((pair[1], pair[0]) for pair in self.server_data_item_pairs_goal),
            key=lambda x: self._natural_keys(x[0]),
        )
        for pair in goal_server_data:
            print(f"                (saved {pair[0]} {pair[1]})")
        print("           )")
        print("    )")

    def _atoi(self, text):
        return int(text) if text.isdigit() else text

    def _natural_keys(self, text):
        return [self._atoi(c) for c in re.split(r"(\d+)", text)]


import argparse

parser = argparse.ArgumentParser()
parser.add_argument("data", type=int, help="the number of data items")
parser.add_argument("layers", type=int, help="the number of layers, must be smaller than the number of data items")
parser.add_argument("scripts", type=int, help="the number of scripts, must be larger or equal than [number of data items]-2")
parser.add_argument("module", choices=["ring-network","small-network","tiny-network"],
                    metavar="module",
                    help='A module describing a network, one of %(choices)s. See example files in the source code directory.')
parser.add_argument("--seed", type=int, help="random seed")

args = parser.parse_args()

generator = DistributedComputing(args.data, args.layers, args.scripts, args.module, args.seed)



#! /usr/bin/env python3

"""Generate a Latex file containing all domain READMEs."""

from pathlib import Path
import sys


DIR = Path(__file__).resolve().parent

HEADER = """\
\\documentclass{article}

\\usepackage[margin=1in]{geometry}
\\usepackage[hidelinks]{hyperref}
\\usepackage[utf8]{inputenc}

\\title{PDDL Generators}

\\author{Jendrik Seipp, \\'Alvaro Torralba, J\\"org Hoffmann}
\\date{}
\\hypersetup{pdfinfo={
Title={PDDL Generators},
Author={Jendrik Seipp, Alvaro Torralba, Joerg Hoffmann}
}}
\\begin{document}
\\maketitle

A collection of PDDL generators, some of which have been used
to generate benchmarks for the International Planning Competition (IPC).

\\tableofcontents
"""

FOOTER = """\
\\end{document}
"""


def get_text(domain_dir: Path):
    files = set(x.name for x in domain_dir.iterdir())
    content = None
    for readme_name in ["README", "README.txt"]:
        if readme_name in files:
            with (domain_dir / readme_name).open() as f:
                content = f.read().strip()
            break
    if content is None:
        sys.exit(f"Error: {domain_dir.name} has no README.txt file.")
    return f"\\section{{{domain_dir.name.capitalize()}}}\n\n\\begin{{verbatim}}\n{content}\\end{{verbatim}}"


def main():
    parts = []
    for domain_dir in sorted(DIR.iterdir()):
        if domain_dir.is_dir() and not domain_dir.name.startswith("."):
            parts.append(get_text(domain_dir))
    print(HEADER + "\n\n".join(parts) + FOOTER)


if __name__ == "__main__":
    main()



import logging
import os
import random
from dataclasses import dataclass
from typing import List, Tuple, Any

import numpy as np
from tarski.grounding import LPGroundingStrategy
from unified_planning.interop import convert_problem_to_tarski
from unified_planning.io import PDDLReader, PDDLWriter
from unified_planning.model import Fluent
from unified_planning.plans import ActionInstance
from unified_planning.shortcuts import SequentialSimulator, BoolType

from amlgym.algorithms.OnlineAlgorithmAdapter import OnlineAlgorithmAdapter
from amlgym.algorithms.SAM import SAM
from amlgym.modeling.trajectory import Trajectory


@dataclass
class RandomAgent(OnlineAlgorithmAdapter):
    """
    A simple baseline for online learning in a fully observable and deterministic
    environment by randomly executing actions. The baselines firstly generates
    a trajectory and then applies the SAM algorithm for offline learning a model
    from the generated trace.

    Example:
        .. code-block:: python

            from unified_planning.io import PDDLReader
            from unified_planning.shortcuts import SequentialSimulator
            from amlgym.algorithms import get_algorithm
            from amlgym.benchmarks import get_domain_path, get_problems_path
            from amlgym.util.util import empty_domain

            domain = 'blocksworld'
            domain_ref_path = get_domain_path(domain)
            input_domain_path = empty_domain(domain_ref_path)
            problem_path = get_problems_path(domain, kind='learning')[0]
            problem = PDDLReader().parse_problem(domain_ref_path, problem_path)

            env = SequentialSimulator(problem=problem)
            baseline = get_algorithm('RandomAgent')
            model, trajectory = baseline.learn(env, input_domain_path)

            print("##################### Learned model #####################")
            print(model)

            print("################# Generated trajectory ##################")
            print(trajectory)

    """
    max_steps: int = 100.

    def learn(self,
              simulator: SequentialSimulator,
              input_domain_path: str,
              seed: int = 123) -> Tuple[str, Trajectory]:
        """
        Learns a PDDL action model from:
         (i)   a simulator of the environment to learn from
         (ii)    a (possibly empty) input model which is required to specify the predicates and operators signature;

        :parameter simulator: environment simulator
        :parameter input_domain_path: input PDDL domain file path
        :parameter seed: random seed for reproducibility

        :return: a string representing the learned PDDL model, and a JSON specification of the trajectory
        """

        # Set seed for reproducibility
        random.seed(seed)
        np.random.seed(seed)

        # Ground actions
        problem_path = 'tmp.pddl'
        PDDLWriter(simulator._problem).write_problem(problem_path)
        ground_actions = self._ground_actions(input_domain_path, problem_path)
        os.remove(problem_path)

        # Get initial state
        state = simulator.get_initial_state()

        trace_actions = []
        trace_states = [state]

        for i in range(self.max_steps):

            action_label = random.choice(ground_actions)
            operator = simulator._problem.action(action_label[0])
            args = [simulator._problem.object(o) for o in action_label[1]]
            action = ActionInstance(operator, tuple(args))

            next_state = simulator.apply(state, action)

            if next_state is not None:
                state = next_state

            trace_states.append(next_state)
            trace_actions.append(action)

        # Store generated trajectory by filtering out failed actions
        trajectory_path = 'tmp_trajectory'
        success_states = [s for s in trace_states if s is not None]
        success_actions = [
            a for s, a in zip(trace_states[1:], trace_actions)
            if s is not None
        ]
        trajectory = Trajectory(success_states, success_actions)
        trajectory.write(trajectory_path)

        model = SAM().learn(input_domain_path, [trajectory_path])

        return model, Trajectory(trace_states, trace_actions)

    def _ground_actions(self, input_domain_path: str, problem_path: str) -> List[Any]:

        # Initialize actions grounder with tarski
        _tmp_problem = PDDLReader().parse_problem(input_domain_path, problem_path)
        # Add a dummy fluent to show `preconditions:` and `effects:` sections in the PDDL file
        dummy_fluent = Fluent('dummy', BoolType())
        if dummy_fluent not in _tmp_problem.fluents:
            _tmp_problem.add_fluent(dummy_fluent)
        _tmp_problem.set_initial_value(dummy_fluent, True)
        # Rebuild actions with no preconditions/effects
        for action in _tmp_problem.actions:
            action.clear_preconditions()
            action.clear_effects()
            # ensure `preconditions:` and `effects:` sections in the PDDL file
            action.add_precondition(dummy_fluent)
            action.add_effect(dummy_fluent, True)
        # Remove problem goal to avoid tarski reachability issues
        _tmp_problem.clear_goals()

        tarski_problem = convert_problem_to_tarski(_tmp_problem)
        grounder = LPGroundingStrategy(tarski_problem)

        logging.debug("Grounding actions with tarski...")
        ground_actions = grounder.ground_actions()

        ground_action_labels = list()
        for op_name, param_combos in ground_actions.items():
            for args in param_combos:
                ground_action_labels.append((op_name, args))

        return ground_action_labels



from subprocess import Popen, PIPE
import numpy as np
import numpy.random as npr
import random
import torch
import matplotlib.pyplot as plt
import torchvision.datasets as datasets
import math
import argparse
import os


obj_num = 5
state_num = 800
mnist_dataset, target_index = None, None


def const_init_states(generator_addr):
    init_states = []

    p = Popen(f"{generator_addr} -n {obj_num} -p {state_num}", stdout=PIPE, shell=True)
    line = p.stdout.readline()
    for _ in range(state_num):
        init_states.append(p.stdout.readline().decode("utf-8")[:-1])
        p.stdout.readline()
    return init_states


def init_mnist():
    global mnist_dataset, target_index
    mnist_dataset = datasets.MNIST(root='./data', train=True, download=True, transform=None)
    targets = mnist_dataset.targets
    target_index = {}
    for i in range(10):
        target_index[i] = np.argwhere(targets==i).flatten().tolist()


'''
Maintain 3 things: a PDDL model, an numpy array representing image and the current image
'''


def add_to_arr(block, arr, grounds, remains, state_str):
    below = int(state_str[2*block-1])
    if below==0:
        ground = random.choice(grounds)
        grounds.remove(ground)
        arr[obj_num, ground] = block
    else:
        if below in remains:
            remains.remove(below)
            add_to_arr(below, arr, grounds, remains, state_str)
        pos = np.where(arr==below)
        arr[pos[0]-1, pos[1]] = block


def convert_str_arr(state_str):
    arr = np.zeros((obj_num+1, obj_num))
    grounds = list(range(obj_num))
    remains = list(range(1, obj_num+1))
    while remains:
        block = remains.pop(0)
        add_to_arr(block, arr, grounds, remains, state_str)
    return arr


def convert_str_model(state_str):
    model = {f'clear {i}' for i in range(1, obj_num+1)}
    model.add("arm-empty")
    for block in range(1, obj_num+1):
        below = int(state_str[2*block-1])
        if below == 0:
            model.add(f'on-table {block}')
        else:
            model.add(f'on {block} {below}')
            model.remove(f'clear {below}')
    return model

def get_image_from_array(arr):
    '''
    Randomly generate corresponding image
    arr should be a (obj_num+1, obj_num) numpy array (with top row for arm)
    Return a [obj_num+1, obj_num, 28, 28] shape array
    '''
    result = [[] for i in range(obj_num+1)]
    for i in range(obj_num+1):
        for j in range(obj_num):
            num = arr[i][j]
            result[i].append(np.asarray(mnist_dataset.__getitem__(npr.choice(target_index[num]))[0]))
    return np.array(result)


'''
Random Action, Get Next Model, Array and Image
'''


pickup = {
    "precondition": lambda obj: {f'clear {obj}', f'on-table {obj}', 'arm-empty'},
    "add_eff": lambda obj: {f'holding {obj}'},
    "del_eff": lambda obj: {f'clear {obj}', f'on-table {obj}', 'arm-empty'}
}

putdown = {
    "precondition": lambda obj: {f'holding {obj}'},
    "add_eff": lambda obj: {f'clear {obj}', f'on-table {obj}', 'arm-empty'},
    "del_eff": lambda obj: {f'holding {obj}'}
}

stack = {
    "precondition": lambda obj, underobj: {f'clear {underobj}', f'holding {obj}'},
    "add_eff": lambda obj, underobj: {f'clear {obj}', f'on {obj} {underobj}', 'arm-empty'},
    "del_eff": lambda obj, underobj: {f'clear {underobj}', f'holding {obj}'}
}

unstack = {
    "precondition": lambda obj, underobj: {f'clear {obj}', f'on {obj} {underobj}', 'arm-empty'},
    "add_eff": lambda obj, underobj: {f'clear {underobj}', f'holding {obj}'},
    "del_eff": lambda obj, underobj: {f'clear {obj}', f'on {obj} {underobj}', 'arm-empty'},
}


def random_action(current_model):
    actions = []
    for block in range(1, obj_num+1):
        if pickup["precondition"](block).issubset(current_model):
            actions.append(("pickup", block))
        if putdown["precondition"](block).issubset(current_model):
            actions.append(("putdown", block))
        for below in range(1, obj_num+1):
            if stack["precondition"](block, below).issubset(current_model):
                actions.append(("stack", block, below))
            if unstack["precondition"](block, below).issubset(current_model):
                actions.append(("unstack", block, below))
    return random.choice(actions)


def get_next_model(current_model, action):
    if action[0]=='pickup':
        next_model = (current_model-pickup['del_eff'](action[1])).union(pickup['add_eff'](action[1]))
    elif action[0]=='putdown':
        next_model = (current_model-putdown['del_eff'](action[1])).union(putdown['add_eff'](action[1]))
    elif action[0]=='stack':
        next_model = (current_model-stack['del_eff'](action[1], action[2])).union(stack['add_eff'](action[1], action[2]))
    elif action[0]=='unstack':
        next_model = (current_model - unstack['del_eff'](action[1], action[2])).union(unstack['add_eff'](action[1], action[2]))
    return next_model


def get_next_arr_img(current_arr, current_img, action):
    next_arr = current_arr.copy()
    next_img = current_img.copy()

    # Arm position is in the middle of the top row
    arm_pos = (0, math.floor(obj_num/2))
    random_zero = np.asarray(mnist_dataset.__getitem__(npr.choice(target_index[0]))[0])
    block_pos = np.where(current_arr==action[1])

    if action[0]=='pickup' or action[0]=='unstack':
        next_arr[arm_pos[0], arm_pos[1]] = action[1]
        next_arr[block_pos[0], block_pos[1]]=0
        next_img[arm_pos[0], arm_pos[1]] = next_img[block_pos[0], block_pos[1]]
        next_img[block_pos[0], block_pos[1]] = random_zero
    elif action[0]=='stack':
        below_pos = np.where(current_arr==action[2])
        next_arr[below_pos[0]-1, below_pos[1]] = action[1]
        next_arr[arm_pos[0], arm_pos[1]] = 0
        next_img[below_pos[0]-1, below_pos[1]] = next_img[arm_pos[0], arm_pos[1]]
        next_img[arm_pos[0], arm_pos[1]] = random_zero
    elif action[0]=='putdown':
        ground = np.random.choice(np.where(current_arr[-1]==0)[0])
        next_arr[obj_num, ground] = action[1]
        next_arr[arm_pos[0], arm_pos[1]] = 0
        next_img[obj_num, ground] = next_img[arm_pos[0], arm_pos[1]]
        next_img[arm_pos[0], arm_pos[1]] = random_zero

    return next_arr, next_img


def show_mnist_image(raw):
    plt.figure()
    digits = np.concatenate(np.concatenate(raw,axis=1), axis=1).astype(np.uint8)
    plt.imshow(255-digits, cmap='gray')


'''
Create Dataset
'''


def get_actions_and_props():
    actions = [f'pickup {block}' for block in range(1, obj_num+1)]
    actions.extend([f'putdown {block}' for block in range(1, obj_num+1)])
    actions.extend([f'stack {block} {below}' for block in range (1, obj_num+1) for below in range(1, obj_num+1) if block!=below])
    actions.extend([f'unstack {block} {below}' for block in range (1, obj_num+1) for below in range(1, obj_num+1) if block!=below])
    actions = {k: v for v, k in enumerate(actions)}
    propositions = ['arm-empty']
    propositions.extend([f'clear {block}' for block in range(1, obj_num+1)])
    propositions.extend([f'on-table {block}' for block in range(1, obj_num+1)])
    propositions.extend([f'holding {block}' for block in range(1, obj_num+1)])
    propositions.extend([f'on {block1} {block2}' for block1 in range(1, obj_num+1) for block2 in range(1, obj_num+1) if block1!=block2])
    propositions = {k: v for v, k in enumerate(propositions)}
    return actions, propositions


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="-o objectnum, -t trace_num, -l trace_length, -s save_addr")
    parser.add_argument("-o", type=int, default=5, help="object num")
    parser.add_argument("-t", type=int, default=800, help="trace num")
    parser.add_argument("-l", type=int, default=10, help="trace length")
    parser.add_argument("-s", default="data", help="save address")
    parser.add_argument("-g", default="blocks-world-generator-and-planner/bbwstates_src/bbwstates", help="initial state generator")
    parser.add_argument("--seed", type=int, default=8800)
    args = parser.parse_args()

    torch.manual_seed(args.seed)
    random.seed(args.seed)
    npr.seed(args.seed)

    obj_num = args.o
    state_num = args.t
    step_num = args.l
    save_to = args.s
    generator_addr = args.g

    actions, propositions = get_actions_and_props()

    features_img = []
    labels = []
    executed_actions = []

    init_states = const_init_states(generator_addr)
    init_mnist()

    for init_state in init_states:
        print(init_states)
        model = convert_str_model(init_state)
        arr = convert_str_arr(init_state)
        img = get_image_from_array(arr)
        # Randomly break data symmetry
        if random.random()>0.5:
            action = random_action(model)
            model = get_next_model(model, action)
            arr, img = get_next_arr_img(arr, img, action)

        feature = []
        label = []
        executed_action = []
        for i in range(step_num):
            # Record current state img
            feature.append(img)
            # Record current state label
            state_label = np.zeros(len(propositions))
            for p in model:
                state_label[propositions[p]] = 1
            label.append(state_label)
            # Select a random action
            action = random_action(model)
            # Record the action
            executed_action.append(actions[' '.join(map(str, action))])
            # Execute the action
            model = get_next_model(model, action)
            # Get next state
            arr, img = get_next_arr_img(arr, img, action)

        # Record final state label
        state_label = np.zeros(len(propositions))
        for p in model:
            state_label[propositions[p]] = 1
        label.append(state_label)

        features_img.append(feature)
        labels.append(label)
        executed_actions.append(executed_action)

    features_img_tensor = torch.tensor(np.array(features_img))
    labels_tensor = torch.tensor(np.array(labels), dtype=torch.float32)
    actions_tensor = torch.tensor(np.array(executed_actions))

    torch.save(features_img_tensor, os.path.join(save_to, "features_img.pt"))
    torch.save(labels_tensor, os.path.join(save_to, "labels.pt"))
    torch.save(actions_tensor, os.path.join(save_to, "actions.pt"))




from experiment_runner.trace_generator import TraceGenerator
import sys

from experiment_runner.rosame_runner import Rosame_Runner
from rosame_n_sam import OnlyNumericSam
from pddl_plus_parser.lisp_parsers import DomainParser, ProblemParser, TrajectoryParser
from pddl_plus_parser.models import Operator
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data import TensorDataset, DataLoader, random_split
import numpy as np
sys.path.append("/Users/omarwattad/Documents/Action-Model-Research/sam_learning")
from sam_learning.learners.numeric_sam import NumericSAMLearner

class NumericRosame:
    def __init__(self,domain_file, problem_file,traces=None):
        self.domain_file = domain_file
        self.problem_file = problem_file
        self.trace_generator = TraceGenerator(domain_file,problem_file)
        self.algorithm = Rosame_Runner(self.trace_generator, traces)
        self.nsam = None

    def prepare_rosame_data(self, trajectory_path):
        """
        implementation:
        input: should be numeric trajectory
        output: discrete predicates that are the indexes of the predicate from rosame propositions list
        parse domain -> problem  -> trajectory ---> typed predicates/actions
        1. takes the trajectories (file of type *.trajectory)
        2. drops every predicate that is not in the propositions list
        3. for predicate in the s (or s')
            3.1 if the name and all the parameters are matched in one of the propositions of rosame
                3.1.1 put 1 in the index of this predicate
                3.1.2 else 0
        4. do the same for the actions
        5. we have now a triplets of state action state that rosame can learn from
        """
        domain = DomainParser(self.domain_file).parse_domain()
        problem = ProblemParser(self.problem_file, domain).parse_problem()
        observation = TrajectoryParser(domain, problem).parse_trajectory(trajectory_path)
        pre = []
        next = []
        actions = []
        for index, action_triplet in enumerate(observation.components):
            action = action_triplet.grounded_action_call
            prev_state = action_triplet.previous_state.state_predicates
            next_state = action_triplet.next_state.state_predicates
            op = Operator(action=domain.actions[action.name], domain=domain, grounded_action_call=action.parameters)
            actions.append(self.get_action_number(op))
            state = self.get_predicates(prev_state)
            pre.append(state)
            state = self.get_predicates(next_state)
            next.append(state)
        return pre, actions, next


    def get_predicate_index(self,pred): #TODO CHANGE!
        for predicate, index in self.algorithm.rosame.propositions.items():
            predicate_parts = predicate.split()
            predicate_name = predicate_parts[0]
            predicate_params = predicate_parts[2::2]
            if predicate_name == pred.name:
                match = True
                for obj in predicate_params:
                    if obj not in pred.grounded_objects:
                        match = False
                        break
                if match:
                    return index
        return None

    def get_predicates(self,predicates):
        state = [0] * len(self.algorithm.rosame.propositions)
        for _, predicate in predicates.items():
            true_index = [self.get_predicate_index(p) for p in predicate]
            for ind in true_index:
                state[ind] = 1
        return state

    def get_action_number(self,op):
        for action, action_number in self.algorithm.rosame.actions.items():
            action_parts = action.split()
            action_name = action_parts[0]
            action_params = action_parts[2::2]
            if action_name == op.name:
                match = True
                for obj in op.grounded_call_objects:
                    if obj not in action_params:
                        match = False
                        break
                if match:
                    return action_number
        return None

    def learn_rosame(self,trajectory_path):
        """Learns an action model using ROSAME learning."""

        steps_state1, steps_action, steps_state2 = self.prepare_rosame_data(trajectory_path)

        steps_state1_tensor = torch.tensor(np.array(steps_state1)).float()
        steps_action_tensor = torch.tensor(np.array(steps_action))
        steps_state2_tensor = torch.tensor(np.array(steps_state2)).float()

        batch_sz = 1000
        dataset = TensorDataset(steps_state1_tensor, steps_action_tensor, steps_state2_tensor)
        dataloader = DataLoader(dataset, batch_size=batch_sz, shuffle=False)

        parameters = []
        for schema in self.algorithm.rosame.action_schemas:
            parameters.append({'params': schema.parameters(), 'lr': 1e-3})
        optimizer = optim.Adam(parameters)

        for epoch in range(100):
            loss_final = 0.0
            for i, (state_1, executed_actions, state_2) in enumerate(dataloader):
                optimizer.zero_grad()
                precon, addeff, deleff = self.algorithm.rosame.build(executed_actions)
                preds = state_1 * (1 - deleff) + (1 - state_1) * addeff
                loss = F.mse_loss(preds, state_2, reduction='sum')
                validity_constraint = (1 - state_1) * (precon)
                loss += F.mse_loss(validity_constraint,
                                   torch.zeros(validity_constraint.shape, dtype=validity_constraint.dtype),
                                   reduction='sum')
                #     loss += model.constraint_loss()
                loss += 0.2 * F.mse_loss(precon, torch.ones(precon.shape, dtype=precon.dtype), reduction='sum')
                loss.backward()
                optimizer.step()
                loss_final += loss.item() / batch_sz
            if epoch % 10 == 0:
                print('Epoch {} RESULTS: Average loss: {:.10f}'.format(epoch, loss_final))

    def learn_numeric_rosame(self,dir_path):
        """
        After learning rosame get the learned domain and put it as an input for r-n-rosame algorithm along-side the observations!
        Implementation:
        1. get rosame data done!
        2. let rosame learn descrete predicates from this data
        3. take the learned rosame model and send it to R-N-SAM
        4. N-SAM learns the numeric fluents and add them to the same domain of learned rosame
        """
        trace_generator = TraceGenerator(dir_path + "/sailing_domain.pddl", dir_path + "/pfile0.pddl",dir_path)
        partial_domain = DomainParser(dir_path + "/sailing_domain.pddl",partial_parsing=False).parse_domain()
        for num_traj in [1,2,3,4,5,6,7,8,9,10,15]:
            for fold in range(5):
                algorithm = Rosame_Runner(trace_generator, None)
                observations = []
                for num_prob in range(fold * num_traj, (fold + 1) * num_traj):
                    print(num_traj, fold, num_prob)
                    trace_generator = TraceGenerator(dir_path + "/sailing_domain.pddl", dir_path + f"/pfile{num_prob}.pddl",dir_path)
                    algorithm.trace_generator = trace_generator
                    problem = ProblemParser(dir_path + f"/pfile{num_prob}.pddl",domain=partial_domain).parse_problem()
                    observation = TrajectoryParser(partial_domain,problem).parse_trajectory(dir_path + f"/pfile{num_prob}.trajectory")
                    observations.append(observation)
                    algorithm.ground_new_trajectory()
                    algorithm.learn_rosame(observation)
                algorithm.export_rosame_domain(f"/Users/omarwattad/Documents/Action-Model-Research/rosame/numeric_rosame/experiments/sailing/rosame_{num_traj}_{fold}.pddl")
                discrete_domain = DomainParser(f"/Users/omarwattad/Documents/Action-Model-Research/rosame/numeric_rosame/experiments/sailing/rosame_{num_traj}_{fold}.pddl",partial_parsing=False).parse_domain()
                nrosame = OnlyNumericSam(partial_domain=discrete_domain, clean_data=False)
                learned_model , _ = nrosame.learn_action_model(observations)
                with open(f"/Users/omarwattad/Documents/Action-Model-Research/rosame/numeric_rosame/experiments/sailing/nRosame_{num_traj}_{fold}.pddl","w") as f:
                    f.write(learned_model.to_pddl())

                nsam = NumericSAMLearner(partial_domain=partial_domain)
                learned_model , _ = nsam.learn_action_model(observations)
                with open(f"/Users/omarwattad/Documents/Action-Model-Research/rosame/numeric_rosame/experiments/sailing/nsam_{num_traj}_{fold}.pddl","w") as f:
                    f.write(learned_model.to_pddl())

dir_path = "/Users/omarwattad/Documents/Action-Model-Research/numeric_datasets/sailing"
model = NumericRosame(dir_path+"/sailing_domain.pddl",dir_path+"/pfile0.pddl")
model.learn_numeric_rosame(dir_path)



import numpy as np
import numpy.random as npr
import random
import torch
import torchvision
import matplotlib.pyplot as plt
from macq import generate, extract
import argparse
import os


eminst_dataset, target_index = None, None
palette = {
    'black': np.array([[[0]],[[0]],[[0]]]),
    'red': np.array([[[255]],[[0]],[[0]]])/255,
    'lime': np.array([[[0]],[[255]],[[0]]])/255,
    'blue': np.array([[[0]],[[0]],[[255]]])/255,
    'yellow': np.array([[[255]],[[255]],[[0]]])/255,
    'cyan': np.array([[[0]],[[255]],[[255]]])/255,
    'magenta': np.array([[[255]],[[0]],[[255]]])/255,
}


def init_mnist():
    global eminst_dataset, target_index
    eminst_dataset = torchvision.datasets.EMNIST(root='./data',
                                                 split='balanced',
                                                 train=True,
                                                 download=True,
                                                 transform=torchvision.transforms.Compose([
                                                     lambda img: torchvision.transforms.functional.rotate(img, -90),
                                                     lambda img: torchvision.transforms.functional.hflip(img),
                                                 ]))
    targets = eminst_dataset.targets
    target_index = {}
    # 10 is 'A' for airplane and 29 is 'T' for truck
    for i in [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 29]:
      target_index[i] = np.argwhere(targets==i).flatten().tolist()


def get_class_img(cls, color):
    return np.asarray(eminst_dataset.__getitem__(npr.choice(target_index[cls]))[0])*(1-palette[color])


def state_to_img(state):
    # Turn a macq state to an image
    img = np.empty((grid_shape[0], grid_shape[1], 3, 28, 28))
    
    backgrounds = {(i, j) for i in range(grid_shape[0]) for j in range(grid_shape[1])}
    
    package_images = {}
    # vehicle_locations map each truck/airplane to a location (e.g. city1-2)
    vehicle_locations = {}
    # item_loc map each package/truck/airplane to an (i, j)
    item_loc ={}
    free_loc = {loc:[(r, c) for r in range(i*3, i*3+3) for c in range(3)]
                for i, loc in enumerate(locations)}
    free_loc.update({a:[(r, c) for r in range(i*3, i*3+3) for c in range(3, 6)]
                     for i, a in enumerate(airports)})
    # Record which packages each truck/plane carries
    vehicle_packages = {v:[] for v in trucks+planes}
    
    for truck in trucks:
        # Find truck's location and assign
        for loc in city_locations[truck_city[truck]]:
            if loc in airports:
                fluent_string = f'(at truck {truck} airport {loc})'
            else:
                fluent_string = f'(at truck {truck} location {loc})'
            if state.fluents[all_fluents[fluent_string]]:
                vehicle_locations[truck] = loc
                assigned_pos = random.choice(free_loc[loc])
                item_loc[truck] = assigned_pos
                free_loc[loc].remove(assigned_pos)
                img[assigned_pos] = get_class_img(29, vehicle_color[truck])
                backgrounds.remove(assigned_pos)
                break
                
    for plane in planes:
        # Find plane's location and assign
        # Plane can only be at airports
        for airport in airports:
            if state.fluents[all_fluents[f'(at airplane {plane} airport {airport})']]:
                vehicle_locations[plane] = airport
                assigned_pos = random.choice(free_loc[airport])
                item_loc[plane] = assigned_pos
                free_loc[airport].remove(assigned_pos)
                img[assigned_pos] = get_class_img(10, vehicle_color[plane])
                backgrounds.remove(assigned_pos)
                break
                
    for i, package in enumerate(packages, 1):
        # Give package an image
        # We remember this image as the original black image for package
        # As it is hard to recover black image from the color image
        package_images[package] = get_class_img(i, 'black')
        # If package is free
        for loc in locations+airports:
            if loc in airports:
                fluent_string = f'(at obj {package} airport {loc})'
            else:
                fluent_string = f'(at obj {package} location {loc})'
            if state.fluents[all_fluents[fluent_string]]:
                assigned_pos = random.choice(free_loc[loc])
                item_loc[package] = assigned_pos
                free_loc[loc].remove(assigned_pos)
                img[assigned_pos] = package_images[package]
                backgrounds.remove(assigned_pos)
                break
        # If package is in a truck or plane
        for vehicle in trucks+planes:
            if vehicle in trucks:
                fluent_string = f'(in obj {package} truck {vehicle})'
            else:
                fluent_string = f'(in obj {package} airplane {vehicle})'
            if state.fluents[all_fluents[fluent_string]]:
                vehicle_packages[vehicle].append(package)
                assigned_pos = random.choice(free_loc[vehicle_locations[vehicle]])
                item_loc[package] = assigned_pos
                free_loc[vehicle_locations[vehicle]].remove(assigned_pos)
                packge_img_wrt_vehicle = package_images[package] * (1-palette[vehicle_color[vehicle]])
                img[assigned_pos] = packge_img_wrt_vehicle
                backgrounds.remove(assigned_pos)
                break
                
    for background in backgrounds:
        img[background] = get_class_img(0, 'black')
                
    return img, (package_images, item_loc, free_loc, vehicle_packages)


def get_next_img(current_img, meta, action):
    next_img = current_img.copy()
    package_images, item_loc, free_loc, vehicle_packages = meta
    
    if action.name=='load-truck' or action.name=='load-airplane':
        # Color the package with truck/airplane color
        package = action.obj_params[0].name
        vehicle = action.obj_params[1].name
        vehicle_packages[vehicle].append(package)
        next_img[item_loc[package]] = package_images[package] * (1-palette[vehicle_color[vehicle]])
    elif action.name=='unload-truck' or action.name=='unload-airplane':
        # Color the package black (replace with the saved original image, easier)
        package = action.obj_params[0].name
        vehicle = action.obj_params[1].name
        vehicle_packages[vehicle].remove(package)
        next_img[item_loc[package]] = package_images[package]
    elif action.name=='drive-truck' or action.name=='fly-airplane':
        # Move the truck/airplane, together with all packages in it
        vehicle = action.obj_params[0].name
        loc_from = action.obj_params[1].name
        loc_to = action.obj_params[2].name
        vehicle_pos_from = item_loc[vehicle]
        vehicle_pos_to = random.choice(free_loc[loc_to])
        free_loc[loc_from].append(vehicle_pos_from)
        free_loc[loc_to].remove(vehicle_pos_to)
        item_loc[vehicle] = vehicle_pos_to
        # Move vehicle
        temp = next_img[vehicle_pos_to].copy()
        next_img[vehicle_pos_to] = next_img[vehicle_pos_from]
        next_img[vehicle_pos_from] = temp
        # Move all the packages the vehicle carries
        for package in vehicle_packages[vehicle]:
            package_pos_from = item_loc[package]
            package_pos_to = random.choice(free_loc[loc_to])
            free_loc[loc_from].append(package_pos_from)
            free_loc[loc_to].remove(package_pos_to)
            item_loc[package] = package_pos_to
            temp = next_img[package_pos_to].copy()
            next_img[package_pos_to] = next_img[package_pos_from]
            next_img[package_pos_from] = temp
            
    return next_img, meta


def show_mnist_image(raw):
    plt.figure()
    img_shown = np.transpose(raw, (0, 1, 3, 4, 2))
    img_shown = np.concatenate(np.concatenate(img_shown,axis=1), axis=1).astype(np.uint8)

    plt.imshow(255-img_shown)


def get_actions_and_props(packages, trucks, planes, locations, airports, cities):
    actions = [f"load-truck location {loc} obj {package} truck {truck}"
    for loc in locations+airports for package in packages for truck in trucks]
    actions.extend([f"load-airplane airplane {plane} airport {airport} obj {package}"
        for plane in planes for airport in airports for package in packages])
    actions.extend([f"unload-truck location {loc} obj {package} truck {truck}"
        for loc in locations+airports for package in packages for truck in trucks])
    actions.extend([f"unload-airplane airplane {plane} airport {airport} obj {package}"
        for plane in planes for airport in airports for package in packages])
    actions.extend([f"drive-truck city {city} location {loc1} location {loc2} truck {truck}"
        for city in cities for loc1 in locations+airports for loc2 in locations+airports if loc1!=loc2 for truck in trucks])
    actions.extend([f"fly-airplane airplane {plane} airport {airport1} airport {airport2}"
        for plane in planes for airport1 in airports for airport2 in airports if airport1!= airport2])
    actions = {k: v for v, k in enumerate(actions)}
    propositions = [f"at location {loc} movable {movable}"
    for loc in locations+airports for movable in packages+trucks+planes]
    propositions.extend([f"in obj {package} transport {transport}"
        for package in packages for transport in trucks+planes])
    propositions.extend([f"in-city city {city} location {loc}"
        for city in cities for loc in locations+airports])
    propositions = {k: v for v, k in enumerate(propositions)}
    return actions, propositions


def state_to_label(state):
    label = np.zeros(len(propositions))
    for f in state.fluents:
        serialized_list = f._serialize()[1:-1].split(' ')
        if f.name=='at':
            f_string = f'at location {serialized_list[4]} movable {serialized_list[2]}'
        elif f.name=='in':
            f_string = f'in obj {serialized_list[2]} transport {serialized_list[4]}'
        elif f.name=='in-city':
            f_string = f'in-city city {serialized_list[4]} location {serialized_list[2]}'
        label[propositions[f_string]] = state.fluents[f]
    # macq does not recognise static propositions as fluents
    label[propositions["in-city city city1 location city1-1"]] = 1
    label[propositions["in-city city city1 location city1-2"]] = 1
    label[propositions["in-city city city2 location city2-1"]] = 1
    label[propositions["in-city city city2 location city2-2"]] = 1
    return label


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-t", type=int, default=3000, help="trace num")
    parser.add_argument("-l", type=int, default=10, help="trace length")
    parser.add_argument("--num_packages", type=int, default=6, help="number of packages")
    parser.add_argument("--skip", type=int, default=1, help="skip between traces")
    parser.add_argument("-s", default="data", help="save address")
    parser.add_argument("--pddl_dom", default="data/pddl/logistics/domain.pddl")
    parser.add_argument("--pddl_prob", default="data/pddl/logistics/prob01.pddl")
    parser.add_argument("--seed", type=int, default=8800)
    args = parser.parse_args()

    torch.manual_seed(args.seed)
    random.seed(args.seed)
    npr.seed(args.seed)

    # Hardcode for now
    grid_shape = (6, 6)
    trucks = ['truckred', 'trucklime']
    planes = ['planeblue', 'planeyellow']
    city_locations = {'city1':['city1-1', 'city1-2'],
                      'city2':['city2-1', 'city2-2']}
    truck_city = {'truckred':'city1', 'trucklime':'city2'}
    locations = ['city1-1', 'city2-1']
    airports = ['city1-2', 'city2-2']
    num_packages = args.num_packages
    packages = [f'package{i}' for i in range(1, num_packages+1)]

    vehicle_color = {f'truck{c}':c for c in ['red', 'lime']}
    vehicle_color.update({f'plane{c}':c for c in ['blue', 'yellow']})

    trace = generate.pddl.VanillaSampling(dom=args.pddl_dom,
                                      prob=args.pddl_prob,
                                      plan_len = args.t*(args.l+args.skip), num_traces = 1, max_time=120).traces[0]
    all_fluents = {f._serialize():f for f in trace.fluents}

    init_mnist()
    actions, propositions = get_actions_and_props(packages, trucks, planes, locations, airports, list(city_locations.keys()))

    traces_images = []
    traces_actions = []
    traces_labels = []

    for n in range(args.t):
        trace_images = []
        trace_actions = []
        trace_labels = []
        img, meta = state_to_img(trace.steps[n*(args.l+args.skip)].state)
        for step in range(n*(args.l+args.skip), n*(args.l+args.skip)+args.l):
            trace_images.append(img)
            trace_labels.append(state_to_label(trace.steps[step].state))
            # action parameter order may not be the same as our model
            action = trace.steps[step].action
            action_obj_params = []
            for o in action.obj_params:
                if 'airplane' not in action.name and o.obj_type=='airport':
                    o.obj_type = 'location'
                action_obj_params.append(o)
            action_obj_params = sorted(action_obj_params, key=lambda o:o.obj_type)
            trace_actions.append(actions[f"{action.name} {' '.join([o.details()for o in action_obj_params])}"])
            img, meta = get_next_img(img, meta, trace.steps[step].action)
        
        # Record final state label
        trace_labels.append(state_to_label(trace.steps[n*(args.l+args.skip)+args.l].state))
        
        traces_images.append(trace_images)
        traces_actions.append(trace_actions)
        traces_labels.append(trace_labels)


    traces_images_tensor = torch.tensor(np.array(traces_images), dtype=torch.float32)
    traces_actions_tensor = torch.tensor(np.array(traces_actions))
    traces_labels_tensor = torch.tensor(np.array(traces_labels), dtype=torch.float32)

    torch.save(traces_images_tensor, os.path.join(args.s, "features_img.pt"))
    torch.save(traces_actions_tensor, os.path.join(args.s, "actions.pt"))
    torch.save(traces_labels_tensor, os.path.join(args.s, "labels.pt"))


import copy
import itertools
import os
import re
from collections import defaultdict


class Operator:

    def __init__(self, operator_name, parameters, precs_pos, precs_neg, eff_pos, eff_neg):

        self.operator_name = operator_name
        self.parameters = parameters
        self.precs_pos = precs_pos
        self.precs_neg = precs_neg
        self.eff_pos = eff_pos
        self.eff_neg = eff_neg

    def __str__(self):
        return f"{self.operator_name}({','.join(self.parameters)})"


class SimpleDomainReader:

    def __init__(self, input_file=None):
        self.input_file = input_file
        self.types_hierarchy = None
        self.operators = None
        self.predicates = None
        self.ground_actions = defaultdict(list)
        self.ground_action_labels = set()

        if input_file is not None:
            self.clean_pddl_domain_file(input_file)
            self.read(f'{input_file}_clean')
            os.remove(f'{input_file}_clean')

    def __str__(self):
        return "\n\n".join(self.operators)

    def read(self, f_name):
        self.types_hierarchy = self.read_object_types_hierarchy(f_name)
        self.operators = self.read_operators(f_name)
        self.predicates = self.read_predicates(f_name)

    def empty(self):
        for o in self.operators:
            o.precs_pos = set()
            o.eff_pos = set()
            o.eff_neg = set()

        if self.ground_actions is not None:
            for op_name, op_actions in self.ground_actions.items():
                for a in op_actions:
                    a.precs_pos = set()
                    a.eff_pos = set()
                    a.eff_neg = set()

    def init_prec_eff(self):

        for operator in self.operators:
            # if operator.precs is None:
            print(f'[Info] Initializing uncertain preconditions of {operator.operator_name} to the maximal superset of preconditions.')

            preconditions_superset = self.get_op_relevant_predicates(operator)

            # Format preconditions syntax
            precs_uncert = {f"{p[1:-1].split()[0]}({','.join(p[1:-1].split()[1:])})"
                            if len(p[1:-1].split()) > 1 else f"{p[1:-1].split()[0]}()" for p in preconditions_superset}
            # operator.precs = {p: 1.0 for p in precs_uncert}
            operator.precs_pos = precs_uncert

            # if operator.eff_pos is None:
            print(f'[Info] Initializing uncertain positive effects of {operator.operator_name} to the maximal superset of possible effects.')

            eff_pos_superset = self.get_op_relevant_predicates(operator)

            # Format preconditions syntax
            eff_pos_superset = {f"{p[1:-1].split()[0]}({','.join(p[1:-1].split()[1:])})"
                                if len(p[1:-1].split()) > 1 else f"{p[1:-1].split()[0]}()" for p in eff_pos_superset}
            # operator.eff_pos = {p: 0. for p in eff_pos_superset}
            operator.eff_pos = eff_pos_superset

            # if operator.eff_neg is None:
            print(f'[Info] Initializing uncertain negative effects of {operator.operator_name} to the maximal superset of possible effects.')

            eff_neg_superset = self.get_op_relevant_predicates(operator)

            # Format preconditions syntax
            eff_neg_superset = {f"{p[1:-1].split()[0]}({','.join(p[1:-1].split()[1:])})"
                                if len(p[1:-1].split()) > 1 else f"{p[1:-1].split()[0]}()" for p in eff_neg_superset}
            # operator.eff_neg_uncert = {p: 0. for p in eff_neg_superset}
            operator.eff_neg = eff_neg_superset

    def read_object_types_hierarchy(self, f_name):
        with open(f_name, 'r') as f:
            data = f.read().split("\n")

            objects_row = [el.replace(")", "").strip()
                           for el in re.findall(r":types.*\(:predicates","++".join(data))[0].replace(":types","").replace("(:predicates", "").split("++")
                           if el.strip() != ""]

            objects = defaultdict(list)
            obj_of_same_type = []

            for row in objects_row:
                row = row.replace("(", "").replace(")", "")
                if row.find("- ") != -1:
                    [objects['objects'].append(el) for el in row.strip().split("- ")[0].split()]
                    [objects['objects'].append(el) for el in row.strip().split("- ")[1].split()]
                    objects[row.strip().split("- ")[1].strip()].extend([el.strip()
                                                                        for el in row.strip().split("- ")[0].strip().split()]
                                                                       + obj_of_same_type
                                                                       + [row.strip().split("- ")[1].strip()])
                    obj_of_same_type = []
                else:
                    [objects['objects'].append(el) for el in row.split()]
                    [obj_of_same_type.append(el) for el in row.split()]

            for object_key, object_values in objects.items():
                if object_key != 'objects':

                    for val in object_values:

                        for key in objects.keys():
                            if val == key:
                                objects[object_key] = [el for el in objects[object_key] + objects[val]]

            for key in objects.keys():
                objects[key] = list(set(objects[key]))

        return objects

    def read_operators(self, f_name):

        operators = dict()

        with open(f_name, "r") as f:
            data = [el.strip().lower() for el in f.read().split("\n")]
            all_action_schema = " ".join(data)[" ".join(data).index(":action"):]

            # Read certain operator preconditions and effects
            operators_cert = [o.strip().lower() for o in re.findall("action(.*?) :parameters", all_action_schema)
                              if not o.strip().lower().endswith('-uncert')]
            for operator_name in operators_cert:

                # Read operator parameters
                action_schema = re.findall(":action {}(.*?)(?:action|$)".format(operator_name), all_action_schema)[0]
                op_params_row = re.findall(":parameters(.*?):precondition", action_schema)[0].strip()[1:-1]
                params = [p.strip() for p in op_params_row.split() if p.strip() != '-']
                op_params = dict()

                params_of_type = []
                for el in params:
                    if '?' in el:
                        params_of_type.append(el)
                    else:
                        for p in params_of_type:
                            op_params[p] = el
                            params_of_type = []

                # Read operator certain preconditions
                # op_precs_row = re.findall(":precondition(.*?):effect", action_schema)[0].strip()[1:-1]
                op_precs_row = re.findall(":precondition(.*?):effect", action_schema)[0]
                op_precs_row = op_precs_row.replace('( not ', '(not ').replace(') )', '))').replace('  ', ' ').replace('( ', '(')
                precs_cert_neg = {e.strip()[1:-1].replace('not', '', 1).strip() for e in re.findall(r"\(not [^)]*\)\)", op_precs_row)
                                  if not len(e.replace('(and', '').replace(')', '').strip()) == 0}
                precs_cert_pos = {p.strip() for p in re.findall(r'(?<!\(not\s)\([^()]*\)', op_precs_row)
                                  if not len(p.replace('(and', '').replace(')', '').strip()) == 0}

                # Read operator certain effects
                op_effects_row = re.findall(":effect(.*?)(?:action|$)", action_schema)[0]
                op_effects_row = op_effects_row.replace('( not ', '(not ').replace(') )', '))').replace('  ', ' ').replace('( ', '(')
                eff_neg_cert = {e.strip()[1:-1].replace('not', '', 1).strip() for e in re.findall(r"\(not[^)]*\)\)", op_effects_row)
                                  if not len(e.replace('(and', '').replace(')', '').strip()) == 0}
                eff_pos_cert = {e.strip() for e in re.findall(r'(?<!\(not\s)\([^()]*\)', op_effects_row)
                                if not len(e.replace('(and', '').replace(')', '').strip()) == 0}

                # Format preconditions and effects syntax
                precs_cert_pos = {f"{p[1:-1].split()[0]}({','.join(p[1:-1].split()[1:])})"
                         if len(p[1:-1].split()) > 1 else f"{p[1:-1].split()[0]}()" for p in precs_cert_pos}
                precs_cert_neg = {f"{p[1:-1].split()[0]}({','.join(p[1:-1].split()[1:])})"
                         if len(p[1:-1].split()) > 1 else f"{p[1:-1].split()[0]}()" for p in precs_cert_neg}
                eff_neg_cert = {f"{p[1:-1].split()[0]}({','.join(p[1:-1].split()[1:])})"
                                  if len(p[1:-1].split()) > 1 else f"{p[1:-1].split()[0]}()" for p in eff_neg_cert}
                eff_pos_cert = {f"{p[1:-1].split()[0]}({','.join(p[1:-1].split()[1:])})"
                                  if len(p[1:-1].split()) > 1 else f"{p[1:-1].split()[0]}()" for p in eff_pos_cert}

                operators[operator_name] = Operator(operator_name, op_params, precs_pos=precs_cert_pos,
                                                    precs_neg=precs_cert_neg,
                                                    eff_pos=eff_pos_cert, eff_neg=eff_neg_cert)

        return list(operators.values())

    def read_predicates(self, f_name):

        with open(f_name, "r") as f:
            data = [el.strip() for el in f.read().split("\n")]
            predicates_row = re.findall(":predicates(.*?):action", " ".join(data))[0]
            predicates_row = [p.strip() for p in re.findall(r"\([^()]*\)", predicates_row)]

            predicates = []
            for p in predicates_row:
                p_name = p[1:-1].split()[0].strip().lower()

                if len(p[1:-1].split()) > 1:
                    p_objs_num = 0
                    p_objs_types = []
                    for s in p[1:-1].split()[1:]:
                        if '?' in s:
                            p_objs_num += 1
                        elif s.strip() != '-':
                            p_objs_types.extend([s]*p_objs_num)
                            p_objs_num = 0
                else:
                    p_objs_types = []

                predicates.append(f"{p_name}({','.join(p_objs_types)})")

        return predicates

    def get_op_relevant_predicates(self, operator):

        op_params = " ".join([f"{k} - {v}" for k, v in operator.parameters.items()])

        obj_type_hierarchy = self.types_hierarchy

        # Get op param types
        single_obj_count = 0
        op_param_types = []
        op_param_supertypes = []
        for el in [el for el in op_params.strip().split() if el.strip() != "-"]:
            if el.startswith("?"):
                single_obj_count += 1
            else:
                [op_param_types.append([el]) if el not in obj_type_hierarchy.keys() else
                 op_param_types.append(obj_type_hierarchy[el])
                 for _ in range(single_obj_count)]

                [op_param_supertypes.append(el) for _ in range(single_obj_count)]
                single_obj_count = 0

        # Get all predicates
        with open(self.input_file, "r") as f:
            data = [el.strip() for el in f.read().split("\n")]
            preds = re.findall(":predicates.+?:action","".join(data))[0]

        all_predicates = sorted(re.findall(r"\([^()]*\)", preds))

        relevant_predicates = []

        for predicate in all_predicates:

            pred_name = predicate.strip()[1:-1].split()[0]

            # Get predicate parameter types
            single_obj_count = 0
            pred_param_types = []
            pred_param_supertypes = []
            for el in [el for el in predicate.strip()[1:-1].strip().split()[1:] if el.strip() != "-"]:
                if el.startswith("?"):
                    single_obj_count += 1
                else:
                    [pred_param_types.append([el]) if el not in obj_type_hierarchy.keys()
                     else pred_param_types.append(obj_type_hierarchy[el])
                     for _ in range(single_obj_count)]

                    [pred_param_supertypes.append(el) for _ in range(single_obj_count)]

                    single_obj_count = 0

            # Check if predicate object types are contained into operator object types
            if all([any([el in [item for sublist in op_param_types for item in sublist]]
                        for el in pred_param_types[i]) for i in range(len(pred_param_types))]):

                all_pred_type_indices = []
                for pred_type in pred_param_types:
                    pred_type_indices = ["?param_{}".format(i+1)
                                         for i, op_pred_type in enumerate(op_param_types)
                                         if len([el for el in pred_type if el in op_pred_type]) > 0]
                                         # if op_pred_type == pred_type]
                    all_pred_type_indices.append(pred_type_indices)

                param_combinations = [list(p) for p in itertools.product(*all_pred_type_indices)]

                # Remove inconsistent combinations according to predicate input types
                param_comb_inconsistent = []
                for comb in param_combinations:

                    comb_param_types = []
                    for param in comb:
                        comb_param_types.append(op_param_supertypes[int(param.split("_")[1]) - 1])

                    for k, op_param_type in enumerate(comb_param_types):

                        if not ((pred_param_supertypes[k] in obj_type_hierarchy.keys() \
                                 and op_param_type in obj_type_hierarchy[pred_param_supertypes[k]]) \
                                or op_param_type == pred_param_supertypes[k]):

                            param_comb_inconsistent.append(comb)

                            break

                # Remove inconsistent combinations
                [param_combinations.remove(comb) for comb in param_comb_inconsistent]

                if len(all_pred_type_indices) > 0:
                    relevant_predicates.extend(["({} {})".format(pred_name, " ".join(pred_comb))
                                                for pred_comb in param_combinations])
                else:
                    relevant_predicates.extend(["({})".format(pred_name)])

        return sorted(relevant_predicates)

    def write(self, f_name):

        with open(self.input_file, 'r') as f:
            data = [el.strip() for el in f.read().split("\n")]
            domain_name = re.findall(r"domain.+?\)","".join(data))[0].strip()[:-1].split()[-1].strip()

        with open(f_name, 'w') as f:

            # Write domain name and requirements
            f.write(f"(define (domain {domain_name})"
                    f"\n(:requirements)")

            # Write types
            f.write("\n(:types")

            # Remove redundant supertypes, i.e. supertypes in their subtypes
            types_hierarchy = copy.deepcopy(self.types_hierarchy)
            types_hierarchy = {k: [t for t in v if t != k] for k, v in types_hierarchy.items()}

            # Remove redundant subtypes, e.g. if a = sub(b) and b = sub(c), remove a from c explicit
            # subtypes since 'a' is implicitly subtype of 'c'
            for supertype, subtypes in types_hierarchy.items():
                for k, v in types_hierarchy.items():
                    if k != supertype and k != 'objects' and k in subtypes:
                        types_hierarchy[supertype] = [v for v in subtypes if v not in types_hierarchy[k]]

            for supertype, subtypes in types_hierarchy.items():
                subtypes = [t for t in subtypes if t != supertype]
                if supertype != 'objects':
                    f.write("\n\t{} - {}".format('\n\t'.join(subtypes), supertype))
                elif len(self.types_hierarchy.keys()) == 1 and supertype == 'objects':
                    f.write("\n\t{}".format('\n\t'.join(subtypes)))
            f.write("\n)")

            # Write predicates
            f.write("\n(:predicates")
            for p in self.predicates:
                p_name = p.split('(')[0]
                p_types = [t for t in p.split('(')[1][:-1].split(',') if t.strip() != '']
                if len(p_types) > 0:
                    f.write(f"\n\t({p_name} {' '.join([f'?param_{i + 1} - {p_type}' for i, p_type in enumerate(p_types)])})")
                else:
                    f.write(f"\n\t({p_name})")
            f.write("\n)\n\n")

            # Write operators with certain preconditions and effects
            for operator in self.operators:
                f.write(f"\n(:action {operator.operator_name}")
                f.write("\n:parameters ({})".format(" ".join(['{} - {}'.format(param, obj_type) for param, obj_type in operator.parameters.items()])))

                # Format preconditions and effects syntax
                precs_cert = []
                for p in operator.precs_pos:
                    if p.startswith('not(') or p.startswith("(not "):
                        p = p[4:-1]
                        if len([o for o in "(".join(p.split('(')[1:])[:-1].split(',') if o != '']) > 0:
                            precs_cert.append("(not ({} {}))".format(p.split('(')[0], " ".join("(".join(p.split('(')[1:])[:-1].split(','))))
                        else:
                            precs_cert.append(f"(not ({p.split('(')[0].strip()}))")
                    elif len([o for o in "(".join(p.split('(')[1:])[:-1].split(',') if o != '']) > 0:
                        precs_cert.append("({} {})".format(p.split('(')[0], " ".join("(".join(p.split('(')[1:])[:-1].split(','))))
                    else:
                        precs_cert.append(f"({p.split('(')[0].strip()})")

                for p in operator.precs_neg:
                    if len([o for o in "(".join(p.split('(')[1:])[:-1].split(',') if o != '']) > 0:
                        precs_cert.append("(not ({} {}))".format(p.split('(')[0], " ".join("(".join(p.split('(')[1:])[:-1].split(','))))
                    else:
                        precs_cert.append(f"(not ({p.split('(')[0].strip()}))")

                eff_pos_cert = ["({} {})".format(p.split('(')[0], " ".join(p.split('(')[1][:-1].split(',')))
                                if len([o for o in p.split('(')[1][:-1].split(',') if o != '']) > 0 else "({})".format(p.split('(')[0])
                                for p in operator.eff_pos]

                eff_neg_cert = ["(not ({} {}))".format(p.split('(')[0], " ".join(p.split('(')[1][:-1].split(',')))
                                if len([o for o in p.split('(')[1][:-1].split(',') if o != '']) > 0 else "(not ({}))".format(p.split('(')[0])
                                for p in operator.eff_neg]

                f.write("\n:precondition\t(and\n\t{}\n)".format("\n\t".join(precs_cert)))

                f.write("\n:effect\t(and\n\t{}\n)".format("\n\t".join(eff_pos_cert + eff_neg_cert)))
                f.write('\n)\n\n')

            f.write('\n\n)')

    def clean_pddl_domain_file(self, f_name):

        with open(f_name, "r") as f:
            data = [el.lower() for el in f.read().split("\n") if not el.strip().startswith(";")]

        # Remove domain functions
        for i in range(len(data)):

            if data[i].find(":action-costs") != -1:
                data[i] = data[i].replace(":action-costs", "")

            if data[i].find(":functions") != -1:

                for j in range(i, len(data)):

                    if data[j].find(":action") != -1:
                        break
                    else:
                        data[j] = ""

        with open(f"{f_name}_clean", "w") as f:
            [f.write(el.lower() + "\n") for el in data]

        # Rename operator parameters in each action schema
        with open(f"{f_name}_clean", "w") as f:

            all_action_schema = []
            action_indices = []

            for i in range(len(data)):
                row = data[i]

                if row.find(":action ") != -1:
                    action_indices.append(i)

            for i in range(len(action_indices)):

                action_index = action_indices[i]

                if action_index != action_indices[-1]:

                    action_schema = "".join(data[action_index:action_indices[i + 1]])

                else:

                    action_schema = "".join(data[action_index:])

                action_schema = re.sub(' +|\t', ' ', action_schema).replace(":", "\n:").replace("\n:", ":", 1)
                params = [el for i, el in
                          enumerate(re.findall(r"\(.*\)", action_schema.split("\n")[1])[0][1:-1].split())
                          if el.startswith("?")]

                for k, param in enumerate(params):
                    action_schema = action_schema.replace("({} ".format(param), "(?param_++{} ".format(k + 1))
                    action_schema = action_schema.replace(" {} ".format(param), " ?param_++{} ".format(k + 1))
                    action_schema = action_schema.replace(" {})".format(param), " ?param_++{})".format(k + 1))

                action_schema = action_schema.replace("?param_++", " ?param_")

                all_action_schema.append(action_schema)

            for i in range(len(data)):
                if data[i].find(":action ") != -1:
                    break
                f.write("\n" + data[i])

            [f.write("\n\n{}".format(action_schema)) for action_schema in all_action_schema]
            f.write("\n)")



from ._syntactic import syntactic_precision, syntactic_recall
from ._solving import problem_solving
from ._predictive import predictive_power, predicted_effects, applicability


__all__ = [
    "print_metrics",
    "syntactic_precision",
    "syntactic_recall",
    "problem_solving",
    "predictive_power",
    "predicted_effects",
    "applicability",
]


def print_metrics() -> None:
    """
    Display the available metrics.

    :return:
    """
    metrics = [name for name in __all__
               if name not in ["print_metrics", "predictive_power"]]
    print("Available metrics:")
    for m in metrics:
        print(f" - {m}")




#!/usr/bin/env python

import argparse

import itertools
import random


def objectlist(tagtype, n):
    return [tagtype + str(i + 1) for i in range(n)]


def objectSeq(tagtype, n):
    oblist = objectlist(tagtype, n)
    return " ".join(oblist)


def unarypred(symbol, arg):
    return "(" + symbol + " " + arg + ")"


def unaryNumvar(symbol, arg):
    return "(= (" + symbol + " " + arg + ") 0)"


def naryNumvar(symbol, args):
    return "(= (" + symbol + " " + " ".join(args) + ") 0)"


def narypred(symbol, args):
    return "(" + symbol + " " + " ".join(args) + ")"


def factSequence(predicate, tagtype, n, indent="\n     "):
    objlist = objectlist(tagtype, n)
    facts = [unarypred(predicate, ob) for ob in objlist]
    return indent.join(facts)


def numVarSequence(predicate, tagtype, n, indent="\n    "):
    objlist = objectlist(tagtype, n)
    facts = [unaryNumvar(predicate, ob) for ob in objlist]
    return indent.join(facts)


def numVarMatrix(predicate, objlist1, objlist2, indent="\n     "):
    numvars = [
        naryNumvar(predicate, list(obtuple))
        for obtuple in itertools.product(objlist1, objlist2)
    ]
    return indent.join(numvars)


def factChain(pred, tagtype, n, rev=False, indent="\n    ", start0=False):
    objlist = objectlist(tagtype, n)
    si = 1
    if start0:
        objlist = [tagtype + "0"] + objlist
        si = 0

    if rev:
        objlist.reverse()

    facts = list()
    for i in range(len(objlist) - 1):
        facts.append(narypred(pred, objlist[i : i + 2]))
    return indent.join(facts)


# Reading the command line arguments


def get_objects(lstage):
    strobj = "\n"
    strobj += objectSeq("num", num_ints) + " - num \n"
    strobj += objectSeq("stage", last_stage + 1) + " - stage\n"
    strobj += objectSeq("round", num_rounds) + " - round \n"
    strobj += objectSeq("worker", num_workers) + " - worker \n"
    strobj += objectSeq("room", num_workers) + " - room \n"

    return strobj


def get_init():
    indent = "\n    "
    strinit = indent + factChain("NEXT_NUM", "num", num_ints, start0=True)

    seqfacts = []
    for i in range(num_ints - 1):
        seqfacts.append(narypred("NEXT2_NUM", ["num" + str(i), "num" + str(i + 2)]))

    for i in range(1, num_ints + 1):
        for j in range(1, i + 1):
            strinit += (
                indent
                + "(NUM_SUBSTRACT num"
                + str(i)
                + " num"
                + str(j)
                + " num"
                + str(i - j)
                + ")"
            )

    strinit += indent + indent.join(seqfacts)

    strinit += indent + factChain("NEXT_STAGE", "stage", last_stage + 1)
    strinit += indent + factChain("NEXT_ROUND", "round", num_rounds)

    strinit += indent + factChain("NEXT_WORKER", "worker", num_workers, True)
    strinit += indent + "(NEXT_WORKER worker1 noworker)\n"

    for j in range(1, num_rounds + 1):
        if j in normalrounds:
            tround = " tnormal"
        else:
            tround = " tharvest"
        strinit += indent + "(category_round round" + str(j) + tround + ")"

    opencards = [
        "act_labor",
        "act_wood",
        "act_clay",
        "act_reed",
        "act_build",
        "act_plow",
        "act_grain",
        "act_stone",
    ]
    facts = [unarypred("open_action", ob) for ob in opencards]
    strinit += indent + indent.join(facts)

    roundcards = [
        "act_fences",
        "act_sheep",
        "act_sow",
        "act_family",
        "act_improve",
        "act_carrot",
        "act_boar",
        "act_cattle",
    ]
    availcards = opencards + roundcards

    s1cards = roundcards[:4]
    s2cards = roundcards[4:]

    dfacts = []
    random.shuffle(s1cards)
    strinit += indent + "(open_action " + s1cards[0] + ")"
    for k in range(min(4, num_rounds)):
        dfacts.append("(DRAWCARD_ROUND " + s1cards[k] + " round" + str(k + 1) + ")")
    strinit += indent + indent.join(dfacts)

    dfacts = []
    random.shuffle(s2cards)
    for k in range(min(4, num_rounds - 4)):
        dfacts.append("(DRAWCARD_ROUND " + s2cards[k] + " round" + str(k + 5) + ")")
    strinit += indent + indent.join(dfacts)

    for k in range(9, num_rounds + 1):
        strinit += indent + "(DRAWCARD_ROUND void round" + str(k) + ")"

    facts = [unarypred("available_action", ob) for ob in availcards]
    strinit += indent + indent.join(facts)

    for k in range(2, num_workers + 1):
        strinit += indent + "(FOOD_REQUIRED worker" + str(k) + " num" + str(2 * k) + ")"
        strinit += (
            indent + "(FOOD_REQUIRED worker" + str(k) + " num" + str((2 * k) + 1) + ")"
        )

    strinit += indent + "(current_worker  worker2)"
    strinit += indent + "(max_worker  worker2)"
    strinit += indent + "(current_round round1)"
    strinit += indent + "(current_stage stage1)"
    strinit += indent + "(harvest_phase stage1 harvest_init)"

    strinit += indent + "(num_food num" + str(random.randint(0, 3)) + ")"

    for el in ["wood", "clay", "reed", "stone"]:
        strinit += indent + "(SUPPLY_RESOURCE act_" + el + " " + el + ")"
    # for el in ['sheep','boar','cattle']:
    #     strinit += indent + "(SUPPLY_ANIMAL act_"+el+" "+el+")"

    strinit += indent + "(built_rooms room1 worker1)"
    strinit += indent + "(built_rooms room2 worker2)"
    for k in range(3, num_workers + 1):
        strinit += indent + "(space_rooms room" + str(k) + ")"

    cost_list = [4, 6, 15, 30, 60]
    while len(cost_list) < num_workers - 1:
        cost_list.append(cost_list[-1] + 30)

    [250, 210, 180, 150, 120, 90, 60, 30, 15, 6, 4]
    for w, c in zip(range(2, num_workers + 1), cost_list[::-1]):
        strinit += (
            indent + "(= (group_worker_cost worker" + str(w) + ") " + str(c) + ")"
        )

    return strinit


def get_goals(last_stage, must_create_workers):
    strgoals = ""
    indent = "\n    "
    strgoals += indent + "(harvest_phase stage" + str(last_stage) + " harvest_end)"

    if must_create_workers:
        strgoals += indent + "(max_worker worker%d)" % num_workers

    return strgoals


name = "prob_agricola"

normalrounds = [1, 2, 3, 5, 6, 8, 10, 12]
parser = argparse.ArgumentParser()

parser.add_argument("last_stage", type=int)
parser.add_argument("seed", type=int)
parser.add_argument("--num_workers", type=int, default=5)
parser.add_argument("--num_ints", type=int, default=16)
parser.add_argument("--num_rounds", type=int, default=20)
parser.add_argument("--must_create_workers", action="store_true")

args = parser.parse_args()


num_rounds = args.num_rounds
num_workers = args.num_workers
num_ints = max(args.num_ints, 2 + args.num_workers * 2)
last_stage = args.last_stage
random.seed(args.seed)

hrounds = [r for r in range(1, 21) if r not in normalrounds]
num_rounds = hrounds[last_stage - 1] + 1

print("(define (problem " + name + ")")
print(" (:domain agricola)")
print(" (:objects " + get_objects(last_stage) + ")")
print(" (:init " + get_init() + ")")
print(" (:goal (and " + get_goals(last_stage, args.must_create_workers) + "))")
print(" (:metric minimize (total-cost))")
print(")")

# except:
#   print "Usage: " +sys.argv[0] + " <name> "+sys.argv[1]+" <last_stage>"



#! /usr/bin/env python

# TETRIS DOMAIN GENERATOR.
# IPC 2014
# author: Mauro Vallati -- University of Huddersfield

import sys
import random

to_stamp_init = ""
to_stamp_square = ""
to_stamp_straight = ""
to_stamp_L_right = ""
rows = 4


def print_error(msg):
    print(msg, file=sys.stderr)


def help():
    print_error(f"usage: {sys.argv[0]} <grid_size> <block_type> [random_seed]")
    print_error(
        "  grid_size: defines the number of rows. "
        "Only even numbers are accepted. "
        "The number of columns is fixed at 4."
    )
    print_error("  block_type:")
    print_error("    1 -> only 1x1 square blocks")
    print_error("    2 -> only 2x1 blocks")
    print_error("    3 -> only L-shaped blocks")
    print_error("    4 -> mix of blocks")
    sys.exit(2)


def put_squares(real_number_square):
    global to_stamp_square
    global to_stamp_init
    global matrix
    squares = [i for i in range(real_number_square)]
    posizionati = 0
    while posizionati < real_number_square:
        x = random.randint(0, (size_grid // 2))
        y = random.randint(0, rows - 1)
        if matrix[x][y] == "free":
            matrix[x][y] = "square" + str(squares[posizionati])
            to_stamp_square = to_stamp_square + matrix[x][y] + " "
            to_stamp_init = (
                to_stamp_init
                + "(at_square square"
                + str(squares[posizionati])
                + " f"
                + str(x)
                + "-"
                + str(y)
                + "f)"
                + "\n"
            )
            posizionati = posizionati + 1


if len(sys.argv) not in {3, 4}:
    help()
size_grid = int(sys.argv[1])
conf_blocks = int(sys.argv[2])
if len(sys.argv) == 4:
    random.seed(int(sys.argv[3]))
if size_grid < 2:
    print_error("grid_size must be at least 2.")
    help()
if size_grid % 2 != 0:
    print_error("Only even values for grid_size are accepted.")
    help()
if conf_blocks < 1 and conf_blocks > 4:
    print_error("Invalid block configuration. Valid configurations: 1 -- 4.")
    help()

# hey, everything is ok. Let's start to do something.

if size_grid % 2 == 0:
    aggiungi = 0
else:
    aggiungi = 1


matrix = [["free" for i in range(rows)] for j in range(size_grid)]


if conf_blocks == 1:
    # only square 1x1
    max_number_squares = (size_grid // 2) * rows
    put_squares(random.randint(1, max_number_squares))


if conf_blocks == 2:
    # only 2x1
    max_number_straights = (size_grid // 2) * (rows // 2)
    # random... ??
    real_number_straights = random.randint(1, max_number_straights)
    straights = [i for i in range(real_number_straights)]
    posizionati = 0
    while posizionati < real_number_straights:
        x = random.randint(0, (size_grid // 2))
        y = random.randint(0, rows - 1)
        if matrix[x][y] == "free":
            direction = random.randint(1, 4)
            if direction == 1:
                # up
                if x != 0:
                    if matrix[x - 1][y] == "free":
                        matrix[x][y] = "straight" + str(straights[posizionati])
                        matrix[x - 1][y] = "straight" + str(straights[posizionati])
                        to_stamp_straight = to_stamp_straight + matrix[x][y] + " "
                        to_stamp_init = (
                            to_stamp_init
                            + "(at_two straight"
                            + str(straights[posizionati])
                            + " f"
                            + str(x)
                            + "-"
                            + str(y)
                            + "f f"
                            + str(x - 1)
                            + "-"
                            + str(y)
                            + "f)"
                            + "\n"
                        )
                        posizionati = posizionati + 1
                else:
                    direction = 2
            if direction == 2:
                # right
                if y != (rows - 1):
                    if matrix[x][y + 1] == "free":
                        matrix[x][y] = "straight" + str(straights[posizionati])
                        matrix[x][y + 1] = "straight" + str(straights[posizionati])
                        to_stamp_straight = to_stamp_straight + matrix[x][y] + " "
                        to_stamp_init = (
                            to_stamp_init
                            + "(at_two straight"
                            + str(straights[posizionati])
                            + " f"
                            + str(x)
                            + "-"
                            + str(y)
                            + "f f"
                            + str(x)
                            + "-"
                            + str(y + 1)
                            + "f)"
                            + "\n"
                        )
                        posizionati = posizionati + 1
                else:
                    direction = 4
            if direction == 3:
                # down
                if matrix[x + 1][y] == "free":
                    matrix[x][y] = "straight" + str(straights[posizionati])
                    matrix[x + 1][y] = "straight" + str(straights[posizionati])
                    to_stamp_straight = to_stamp_straight + matrix[x][y] + " "
                    to_stamp_init = (
                        to_stamp_init
                        + "(at_two straight"
                        + str(straights[posizionati])
                        + " f"
                        + str(x)
                        + "-"
                        + str(y)
                        + "f f"
                        + str(x + 1)
                        + "-"
                        + str(y)
                        + "f)"
                        + "\n"
                    )
                    posizionati = posizionati + 1
            if direction == 4:
                # left
                if y != 0:
                    if matrix[x][y - 1] == "free":
                        matrix[x][y] = "straight" + str(straights[posizionati])
                        matrix[x][y - 1] = "straight" + str(straights[posizionati])
                        to_stamp_straight = to_stamp_straight + matrix[x][y] + " "
                        to_stamp_init = (
                            to_stamp_init
                            + "(at_two straight"
                            + str(straights[posizionati])
                            + " f"
                            + str(x)
                            + "-"
                            + str(y)
                            + "f f"
                            + str(x)
                            + "-"
                            + str(y - 1)
                            + "f)"
                            + "\n"
                        )
                        posizionati = posizionati + 1

if conf_blocks == 3:
    # only L
    max_number_ls = (size_grid // 4) * (rows // 2)
    # random... ??
    real_number_l_right = random.randint(1, max_number_ls)
    l_rights = [i for i in range(real_number_l_right)]
    posizionati = 0
    while posizionati < real_number_l_right:
        x = random.randint(0, (size_grid // 2) - 1)
        y = random.randint(0, rows - 2)
        if (
            matrix[x][y] == "free"
            and matrix[x + 1][y + 1] == "free"
            and matrix[x + 1][y] == "free"
        ):
            matrix[x][y] = "rightl" + str(l_rights[posizionati])
            matrix[x + 1][y] = "rightl" + str(l_rights[posizionati])
            matrix[x + 1][y + 1] = "rightl" + str(l_rights[posizionati])
            to_stamp_L_right = to_stamp_L_right + matrix[x][y] + " "
            to_stamp_init = (
                to_stamp_init
                + "(at_right_l "
                + matrix[x][y]
                + " f"
                + str(x)
                + "-"
                + str(y)
                + "f f"
                + str(x + 1)
                + "-"
                + str(y)
                + "f f"
                + str(x + 1)
                + "-"
                + str(y + 1)
                + "f)"
                + "\n"
            )
            posizionati = posizionati + 1
if conf_blocks == 4:
    max_number_ls = (size_grid // 4) * (rows // 2)
    real_number_l_right = random.randint(1, max_number_ls)
    l_rights = [i for i in range(real_number_l_right)]
    posizionati = 0
    while posizionati < real_number_l_right:
        x = random.randint(0, (size_grid // 2) - 1)
        y = random.randint(0, rows - 2)
        if (
            matrix[x][y] == "free"
            and matrix[x + 1][y + 1] == "free"
            and matrix[x + 1][y] == "free"
        ):
            matrix[x][y] = "rightl" + str(l_rights[posizionati])
            matrix[x + 1][y] = "rightl" + str(l_rights[posizionati])
            matrix[x + 1][y + 1] = "rightl" + str(l_rights[posizionati])
            to_stamp_L_right = to_stamp_L_right + matrix[x][y] + " "
            to_stamp_init = (
                to_stamp_init
                + "(at_right_l "
                + matrix[x][y]
                + " f"
                + str(x)
                + "-"
                + str(y)
                + "f f"
                + str(x + 1)
                + "-"
                + str(y)
                + "f f"
                + str(x + 1)
                + "-"
                + str(y + 1)
                + "f)"
                + "\n"
            )
            posizionati = posizionati + 1
    # straights
    linea = 1
    posizionati = 0
    while linea < size_grid / 2:
        colonna = 0
        while colonna < rows - 1:
            if (
                matrix[linea][colonna] == "free"
                and matrix[linea - 1][colonna] == "free"
            ):
                if random.randint(1, 4) > 2:
                    matrix[linea][colonna] = "straight" + str(posizionati)
                    matrix[linea - 1][colonna] = "straight" + str(posizionati)
                    to_stamp_straight = to_stamp_straight + matrix[linea][colonna] + " "
                    to_stamp_init = (
                        to_stamp_init
                        + "(at_two straight"
                        + str(posizionati)
                        + " f"
                        + str(linea - 1)
                        + "-"
                        + str(colonna)
                        + "f f"
                        + str(linea)
                        + "-"
                        + str(colonna)
                        + "f)"
                        + "\n"
                    )
                    posizionati = posizionati + 1
            colonna = colonna + 1
        linea = linea + 1
    # squares
    linea = 0
    posizionati = 0
    while linea < size_grid / 2:
        colonna = 0
        while colonna < rows - 1:
            if matrix[linea][colonna] == "free":
                if random.randint(1, 4) > 2:
                    matrix[linea][colonna] = "square" + str(posizionati)
                    to_stamp_square = to_stamp_square + matrix[linea][colonna] + " "
                    to_stamp_init = (
                        to_stamp_init
                        + "(at_square square"
                        + str(posizionati)
                        + " f"
                        + str(linea)
                        + "-"
                        + str(colonna)
                        + "f)"
                        + "\n"
                    )
                    posizionati = posizionati + 1
            colonna = colonna + 1
        linea = linea + 1

    # mix up!
# Finally, time for printing everything out.

print(
    "(define (problem Tetris-"
    + str(size_grid)
    + "-"
    + str(conf_blocks)
    + "-"
    + str(random.randint(0, 9875232))
    + ")"
)
print("(:domain tetris)")
print("(:objects ", end=" ")

for i in range(size_grid):
    print("")
    for j in range(rows):
        x = "f" + str(i) + "-" + str(j) + "f"
        print(x, end=" ")
print("- position")

# eventually, this be commented away. **

if to_stamp_square == "":
    to_stamp_square = "nothing"
if to_stamp_straight == "":
    to_stamp_straight = "nada"
if to_stamp_L_right == "":
    to_stamp_L_right = "nisba"


# until here **

print(to_stamp_square + "- one_square")
print(to_stamp_straight + "- two_straight")
print(to_stamp_L_right + "- right_l")


print(")\n(:init")

# connected cells on the same line
for i in range(size_grid):
    for j in range(rows - 1):
        print(
            "(connected f"
            + str(i)
            + "-"
            + str(j)
            + "f f"
            + str(i)
            + "-"
            + str(j + 1)
            + "f)"
        )
        print(
            "(connected f"
            + str(i)
            + "-"
            + str(j + 1)
            + "f f"
            + str(i)
            + "-"
            + str(j)
            + "f)"
        )

# connected cells on the same column
for i in range(size_grid - 1):
    for j in range(rows):
        print(
            "(connected f"
            + str(i)
            + "-"
            + str(j)
            + "f f"
            + str(i + 1)
            + "-"
            + str(j)
            + "f)"
        )
        print(
            "(connected f"
            + str(i + 1)
            + "-"
            + str(j)
            + "f f"
            + str(i)
            + "-"
            + str(j)
            + "f)"
        )


# free cells
for i in range(size_grid):
    for j in range(rows):
        if matrix[i][j] == "free":
            print("(clear f" + str(i) + "-" + str(j) + "f)")

# occupied cells
print(to_stamp_init, end=" ")
print(")\n(:goal\n(and")

for i in range(0, size_grid // 2 + aggiungi):
    for j in range(rows):
        print("(clear f" + str(i) + "-" + str(j) + "f)")

print(")\n)\n)")


print(";; DESCRIPTION OF THE INITIAL STATE")
for i in range(size_grid):
    print(";; " + str(i) + " ", end=" ")
    for j in range(rows):
        if matrix[i][j] == "free":
            print(" **", end=" ")
        else:
            print(" ##", end=" ")
    print("")



#! /usr/bin/env python

import math
import euclidean_graph
import sys
import random

sys.path.append("../")
sys.path.append("../../")

MAX_SEED = 10000000


def road(f, t, length):
    length = int(math.ceil(length / 10.0))
    print(f"  (road {f} {t})")
    print("  (= (road-length %s %s) %d)" % (f, t, length))


def symmetric_road(f, t, length):
    road(f, t, length)
    road(t, f, length)


if len(sys.argv) != 8:
    raise SystemExit(
        "Usage: <nodes> <size^(1/2)> <degree> <mindistance> <nr-trucks> <nr-packages> <seed>"
    )

nodes = int(sys.argv[1])
size = int(sys.argv[2])
degree = float(sys.argv[3])
epsilon = float(sys.argv[4])
trucks = int(sys.argv[5])
packages = int(sys.argv[6])
seed = float(sys.argv[7])

max_capacity = 4  # maximum number of packages in one truck
assert max_capacity > 2

if not seed:
    seed = random.randrange(MAX_SEED) + 1
random.seed(seed)

#         deg * width * height
# ratio = ---------------------------
#         nodes * pi * Connect^2
connect_distance = math.sqrt((degree * size * size) / (nodes * math.pi * 0.694))

city = euclidean_graph.generate_connected_safe(
    nodes, size, size, connect_distance, epsilon
)

id = (
    "city-sequential-%dnodes-%dsize-%ddegree-%dmindistance-%dtrucks-%dpackages-%dseed"
    % (nodes, size, degree, epsilon, trucks, packages, seed)
)

# city.dump_tikz(open("%s.tex" % id, "w"))

print("; Transport %s" % id)
print()

print("(define (problem transport-%s)" % id)
print(" (:domain transport)")
print(" (:objects")

for i in range(nodes):
    print("  city-loc-%d - location" % (i + 1))
for i in range(trucks):
    print("  truck-%d - vehicle" % (i + 1))

for i in range(packages):
    print("  package-%d - package" % (i + 1))

for i in range(max_capacity + 1):
    print("  capacity-%d - capacity-number" % i)

print(" )")
print(" (:init")

print("  (= (total-cost) 0)")


for i in range(max_capacity):
    print("  (capacity-predecessor capacity-%d capacity-%d)" % (i, i + 1))

for u, v in city.edges:
    print("  ; %d,%d -> %d,%d" % (u.x, u.y, v.x, v.y))

    dist = u.round_distance(v)
    road("city-" + u.name, "city-" + v.name, dist)

packages_position = {}
for i in range(packages):
    packages_position["package-%d" % (i + 1)] = random.choice(city.vertices).name
    print(
        "  (at package-%d city-%s)" % (i + 1, packages_position["package-%d" % (i + 1)])
    )

for i in range(trucks):
    capacity = random.randint(2, max_capacity)
    print("  (at truck-%d city-%s)" % (i + 1, random.choice(city.vertices).name))
    print("  (capacity truck-%d capacity-%d)" % (i + 1, capacity))


print(" )")
print(" (:goal (and")
for i in range(packages):
    target_pos = random.choice(city.vertices).name
    while target_pos == packages_position["package-%d" % (i + 1)]:
        target_pos = random.choice(city.vertices).name
    print("  (at package-%d city-%s)" % (i + 1, target_pos))
print(" ))")

print(" (:metric minimize (total-cost))")
print(")")



#!/usr/bin/env python3

import argparse
from itertools import product
import random
import itertools

from collections import defaultdict
import sys


def parse():
    parser = argparse.ArgumentParser()

    parser.add_argument("x", type=int, help="x scale (minimal 2)")
    parser.add_argument("y", type=int, help="y scale (minimal 2)")
    parser.add_argument(
        "--shapes", type=int, default=2, help="num different key+lock types"
    )
    parser.add_argument("--keys", type=int, default=2, help="number keys vector")
    parser.add_argument("--locks", type=int, default=2, help="number locks vector")
    parser.add_argument(
        "--prob-goal",
        type=float,
        default=1.0,
        help="probability of any key being mentioned in the goal",
    )

    parser.add_argument("--typed", action="store_true", help="Typed domain")

    parser.add_argument("--seed", type=int, default=1, help="random seed")

    parser.add_argument("--output", default=None)

    return parser.parse_args()


def backslash_join(x):
    return "\n       ".join(x)


def adjacent_positions(x, y, positions):
    return [
        adj_pos
        for adj_pos in [(x + 1, y), (x, y + 1), (x - 1, y), (x, y - 1)]
        if adj_pos in positions
    ]


def pos_name(p):
    return f"pos{p[0]}-{p[1]}"


def main():
    args = parse()
    random.seed(args.seed)

    if args.output:
        sys.stdout = open(f"{args.output}/{instance_name}.pddl", "w")

    positions = [(x, y) for x in range(args.x) for y in range(args.y)]
    assert args.locks < len(positions), "There cannot be more locks than positions"
    assert args.shapes <= args.keys, "There cannot be more shapes than keys"
    assert args.shapes <= args.locks, "There cannot be more shapes than locks"
    assert args.prob_goal <= 1.0, "Goal probability cannot be greater than 1.0"
    assert args.prob_goal > 0.0, "Goal probability cannot be 0.0"

    locked_positions = random.sample([p for p in positions], k=args.locks)
    open_positions = [p for p in positions if p not in locked_positions]
    robot_pos = random.choice(open_positions)

    instance_name = f"grid-{args.x}-{args.y}-{args.shapes}-{args.keys}-{args.locks}"
    nodes = [pos_name(p) for p in positions]
    keys = [f"key{k}" for k in range(args.keys)]
    shapes = [f"shape{k}" for k in range(args.shapes)]

    key_shapes = [
        shapes[k] if k < len(shapes) else random.choice(shapes)
        for k in range(args.keys)
    ]
    locked_shapes = [
        shapes[k] if k < len(shapes) else random.choice(shapes)
        for k in range(len(locked_positions))
    ]

    locked_pos_to_shape = {}
    for i, pos in enumerate(locked_positions):
        locked_pos_to_shape[pos] = locked_shapes[i]

    shape_to_keys = defaultdict(list)
    for i, key in enumerate(keys):
        shape_to_keys[key_shapes[i]].append(key)

    key_positions = {}

    i = 0
    reachable_locked = set()
    reachable_locations = [robot_pos]
    unlocked_positions = [p for p in open_positions]

    while len(reachable_locations) < len(positions):
        while i < len(reachable_locations):
            loc = reachable_locations[i]
            adj = adjacent_positions(loc[0], loc[1], positions)
            for pos in adj:
                if pos in reachable_locations:
                    continue

                if pos in unlocked_positions:
                    reachable_locations.append(pos)
                else:
                    reachable_locked.add(pos)
            i += 1

        if len(reachable_locked) > 0:
            pos = random.choice(list(reachable_locked))
            shape = locked_pos_to_shape[pos]

            all_keys_shape = shape_to_keys[shape]
            sure_key = random.choice(all_keys_shape)
            for key in all_keys_shape:
                assert not key in key_positions
                key_positions[key] = (
                    random.choice(reachable_locations)
                    if key == sure_key
                    else random.choice(positions)
                )

            for pos in positions:
                if pos in locked_pos_to_shape and locked_pos_to_shape[pos] == shape:
                    unlocked_positions.append(pos)

            to_remove = []
            for r in reachable_locked:
                if locked_pos_to_shape[r] == shape:
                    to_remove.append(r)
                    reachable_locations.append(r)

            for r in to_remove:
                reachable_locked.remove(r)

    if args.typed:
        place_facts = []
        shape_facts = []
        key_facts = []
    else:
        place_facts = [f"(place {pos_name(pos)})" for pos in positions]
        shape_facts = [f"(shape {shape})" for shape in shapes]
        key_facts = [f"(key {key})" for key in keys]

    conn_facts = [
        f"(conn {pos_name(p1)} {pos_name(p2)})"
        for p1 in positions
        for p2 in adjacent_positions(p1[0], p1[1], positions)
    ]

    key_shape_facts = [
        f"(key-shape  {keys[i]} {key_shapes[i]})" for i in range(args.keys)
    ]

    open_facts = [f"(open {pos_name(pos)})" for pos in open_positions]
    locked_facts = [f"(locked {pos_name(pos)})" for pos in locked_positions]
    lock_shape_facts = [
        f"(lock-shape  {pos_name(locked_positions[i])} {locked_shapes[i]})"
        for i in range(len(locked_positions))
    ]

    key_at_facts = [f"(at {key} {pos_name(key_positions[key])})" for key in keys]

    at_robot_fact = f"(at-robot {pos_name(robot_pos)})"

    goal_facts = []
    for k in range(args.keys):
        if random.random() < args.prob_goal:
            p = random.choice([p for p in positions if p != key_positions[keys[k]]])
            goal_facts.append(f"(at {keys[k]} {pos_name(p)})")

    if not goal_facts:
        k = random.choice(range(args.keys))
        p = random.choice([p for p in positions if p != key_positions[keys[k]]])
        goal_facts.append(f"(at {keys[k]} {pos_name(p)})")

    print(
        f"""(define (problem {instance_name})
    (:domain grid)
    (:objects
    {" ".join(nodes)}
    {" ".join(shapes)}
    {" ".join(keys)}
    )

    (:init
       (arm-empty)

       {backslash_join(place_facts)}
       {backslash_join(shape_facts)}
       {backslash_join(key_facts)}

       {backslash_join(conn_facts)}
       {backslash_join(locked_facts)}
       {backslash_join(lock_shape_facts)}
       {backslash_join(open_facts)}

       {backslash_join(key_shape_facts)}
       {backslash_join(key_at_facts)}
       {at_robot_fact}
    )

    (:goal (and
       {backslash_join(goal_facts)}
    )))
    """
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import argparse
from collections import defaultdict
from z3 import *
from copy import deepcopy


def get_location(x, y):
    return "pos-%d-%d" % (x, y)


def place_block(grid, loc, loc2):
    height = grid[loc[1]][loc[0]]
    new_height = grid[loc2[1]][loc2[0]]

    return "(place-block {} {} n{} n{})".format(
        get_location(*loc), get_location(*loc2), new_height, new_height + 1
    )


def remove_block(grid, loc, loc2):
    height = grid[loc[1]][loc[0]]
    new_height = grid[loc2[1]][loc2[0]]

    return "(remove-block {} {} n{} n{})".format(
        get_location(*loc), get_location(*loc2), new_height, new_height - 1
    )


def move(grid, loc, loc2):
    height = grid[loc[1]][loc[0]]
    new_height = grid[loc2[1]][loc2[0]]

    if height > new_height:
        return "(move-down {} n{} {} n{})".format(
            get_location(*loc), height, get_location(*loc2), new_height
        )
    elif height < new_height:
        return "(move-up {} n{} {} n{})".format(
            get_location(*loc), height, get_location(*loc2), new_height
        )
    else:
        return "(move {} {} n{})".format(
            get_location(*loc), get_location(*loc2), new_height
        )


def print_grid(grid):
    for row in grid:
        print(" ".join(map(str, row)))


def accessible(i_grid, x, y, place_block):
    for (xp, yp) in [(1, 0), (-1, 0), (0, 1), (0, -1)]:
        if (
            x + xp < 0
            or x + xp > len(i_grid[0]) - 1
            or y + yp < 0
            or y + yp > len(i_grid) - 1
        ):
            continue

        if not place_block:
            # The goal must be reached from a location at one height below
            if i_grid[y][x] == i_grid[y + yp][x + xp] + 1:
                return True

        if place_block:
            # The goal must be reached from a location at the same height
            if i_grid[y][x] == i_grid[y + yp][x + xp]:
                return True

    return False


def get_shortest_path(i_grid, ini_loc, goal_loc, place_block, order_grid):
    assert accessible(i_grid, goal_loc[0], goal_loc[1], place_block)
    parent_ptr = {}
    closed, queue = set(), [ini_loc]
    while queue:
        vertex = queue.pop(0)
        if vertex == goal_loc:
            solution = []
            while vertex != ini_loc:
                solution.append(vertex)
                vertex = parent_ptr[vertex]
            solution.append(vertex)
            return solution[::-1]
        if vertex not in closed:
            closed.add(vertex)
            (x, y) = vertex
            for (xp, yp) in [(1, 0), (-1, 0), (0, 1), (0, -1)]:

                if (
                    x + xp < 0
                    or x + xp > len(i_grid[0]) - 1
                    or y + yp < 0
                    or y + yp > len(i_grid) - 1
                ):
                    continue

                if (
                    place_block
                    and abs(i_grid[y][x] - i_grid[y + yp][x + xp]) == 1
                    and (x + xp, y + yp) == goal_loc
                ):
                    continue  # The goal must be reached from a location at the same height

                if (
                    not place_block
                    and i_grid[y][x] - i_grid[y + yp][x + xp] >= 0
                    and (x + xp, y + yp) == goal_loc
                ):
                    continue  # The goal must be reached from a location at one height below

                if (
                    abs(i_grid[y][x] - i_grid[y + yp][x + xp]) <= 1
                    and (x + xp, y + yp) not in parent_ptr
                ):
                    parent_ptr[(x + xp, y + yp)] = (x, y)
                    queue.append((x + xp, y + yp))

    print("Error: no path found")
    print_grid(i_grid)
    print(ini_loc)
    print(goal_loc)
    print(place_block)

    print(closed)

    return None


def solve(
    grid_size_x,
    grid_size_y,
    max_height,
    robots,
    depots,
    i_grid,
    g_grid,
    minimize=True,
    timeout=100000,
):
    i_grid = deepcopy(i_grid)
    # print ("Solving: ",  grid_size_x, grid_size_y, max_height, robots, depots, i_grid, g_grid)
    assert grid_size_y == len(i_grid)

    adjacent = defaultdict(list)

    # A cell X is supporter of another Y if we will use X in order to access Y
    for x in range(grid_size_x):
        for y in range(grid_size_y):
            if x > 0:
                adjacent[(x, y)].append((x - 1, y))

            if y > 0:
                adjacent[(x, y)].append((x, y - 1))

            if x < grid_size_x - 1:
                adjacent[(x, y)].append((x + 1, y))

            if y < grid_size_y - 1:
                adjacent[(x, y)].append((x, y + 1))

    intermediate_board = [
        [Int("intermediate%d-%d" % (c, r)) for c in range(grid_size_x)]
        for r in range(grid_size_y)
    ]
    target_board = [
        [Int("target%d-%d" % (c, r)) for c in range(grid_size_x)]
        for r in range(grid_size_y)
    ]

    order = [
        [Int("order%d-%d" % (c, r)) for c in range(grid_size_x)]
        for r in range(grid_size_y)
    ]
    is_supporter = [
        [Bool("supporter%d-%d" % (c, r)) for c in range(grid_size_x)]
        for r in range(grid_size_y)
    ]
    is_supported = [
        [Bool("supported%d-%d" % (c, r)) for c in range(grid_size_x)]
        for r in range(grid_size_y)
    ]

    supporter_variables = {}
    for x in range(grid_size_x):
        for y in range(grid_size_y):
            for (xp, yp) in adjacent[(x, y)]:
                supporter_variables[(x, y, xp, yp)] = Bool(
                    "supporter%d-%d-%d-%d" % (x, y, xp, yp)
                )

    all_intermediate_board = [cell for column in intermediate_board for cell in column]
    s = Optimize()
    # s.set("timeout", timeout)

    for (y, row) in enumerate(g_grid):
        for (x, tile) in enumerate(row):
            if (x, y) in depots:
                s.add(
                    intermediate_board[y][x] == 0
                )  # The depot location must always have height 0
                s.add(order[y][x] == 0)  # The depot location must always have order 0
            else:
                s.add(order[y][x] > 0)
                s.add(target_board[y][x] == tile)

                s.add(
                    intermediate_board[y][x] <= max_height
                )  # Intermediate_Board cannot have more height than max_height
                s.add(
                    intermediate_board[y][x] >= max(tile, i_grid[y][x])
                )  # Intermediate_Board cannot have negative height

                # All intermediate_board supporter or greater than 0 must have a supporter
                s.add(Implies(intermediate_board[y][x] > 0, is_supported[y][x] == True))

                s.add(
                    is_supported[y][x]
                    == Or(
                        *[
                            supporter_variables[(x, y, xp, yp)] == True
                            for (xp, yp) in adjacent[(x, y)]
                        ]
                    )
                )
                s.add(
                    is_supporter[y][x]
                    == Or(
                        *[
                            supporter_variables[(xp, yp, x, y)] == True
                            for (xp, yp) in adjacent[(x, y)]
                        ]
                    )
                )
                s.add(Implies(is_supporter[y][x], is_supported[y][x]))
                for (xp, yp) in adjacent[(x, y)]:
                    # A cell cannot support its supporter (there is a total order of supporters)
                    s.add(
                        Implies(
                            supporter_variables[(x, y, xp, yp)],
                            order[yp][xp] < order[y][x],
                        )
                    )

                    # The supporter must give access to the cell
                    s.add(
                        Or(
                            supporter_variables[(x, y, xp, yp)] == False,
                            intermediate_board[y][x] == intermediate_board[yp][xp],
                            intermediate_board[y][x] == intermediate_board[yp][xp] + 1,
                        )
                    )

                    # The supporter must give access to remove the cell
                    s.add(
                        Implies(
                            supporter_variables[(x, y, xp, yp)],
                            target_board[y][x] >= target_board[yp][xp],
                        )
                    )

                    # s.add(Or(is_supporter[y][x] == False, is_supported[y][x] == False, is_supporter[yp][xp] == False, is_supported[yp][xp] == False, order[y][x] == order[yp][xp], order[y][x] == order[yp][xp] + 1, order[y][x] == order[yp][xp] - 1))

                    # s.add (intermediate_board[0][0] == 0)
    if minimize:
        h = s.minimize(
            Sum(*all_intermediate_board) * 1000
            + Sum([cell for column in order for cell in column])
        )
    s_result = s.check()
    # print(s)
    # print(s_result)
    if str(s_result).strip() == "unsat":
        # print(s)
        # exit()
        return None
    if str(s_result).strip() == "unknown":
        return None
    m = s.model()

    # for r in range(grid_size_y):
    #     sys.stdout.write ("; ")

    #     for c in range(grid_size_x):
    #         sys.stdout.write (str(m[cells[r][c]])+" ")
    #     print ("")

    target_grid = [
        [int(m[intermediate_board[r][c]].as_long()) for c in range(grid_size_x)]
        for r in range(grid_size_y)
    ]
    order_grid = [
        [int(m[order[r][c]].as_long()) for c in range(grid_size_x)]
        for r in range(grid_size_y)
    ]
    # print ("Init:")
    # print_grid(i_grid)

    # print ("Target:")
    # print_grid(target_grid)

    # print ("Goal:")
    # print_grid(g_grid)

    # print ("Order:")
    # print_grid(order_grid)

    # for x in range(grid_size_x):
    #     for y in range (grid_size_y):
    #         for (xp, yp) in adjacent[(x, y)]:
    #             if m[supporter_variables[(x, y, xp, yp)]]:
    #                 print ((x, y), " is supported by ",  (xp, yp))
    # exit()

    depot_loc = get_location(*depots[0])
    plan = []
    while i_grid != target_grid:
        # Pick lowest position that is different. In case of tie, pick the one highest in order
        different_positions = [
            (x, y)
            for x in range(grid_size_x)
            for y in range(grid_size_y)
            if i_grid[y][x] < target_grid[y][x]
            if accessible(i_grid, x, y, True)
        ]

        if not different_positions:
            print("Error: I cannot place the next position")
            print_grid(i_grid)
            print()
            print_grid(target_grid)
            exit()
        (x, y) = min(
            different_positions,
            key=lambda w: (i_grid[w[1]][w[0]], -order_grid[w[1]][w[0]]),
        )

        plan.append(f"(create-block {depot_loc})")

        path = get_shortest_path(i_grid, depots[0], (x, y), True, order_grid)
        if not path:
            return None
        prev_loc = path[0]
        for loc in path[1:-1]:
            plan.append(move(i_grid, prev_loc, loc))
            prev_loc = loc

        plan.append(place_block(i_grid, path[-2], path[-1]))

        for loc in path[::-1][2:]:
            plan.append(move(i_grid, prev_loc, loc))
            prev_loc = loc

        i_grid[y][x] += 1

    while i_grid != g_grid:

        # Pick highest position that is different
        different_positions = [
            (x, y)
            for x in range(grid_size_x)
            for y in range(grid_size_y)
            if i_grid[y][x] != g_grid[y][x]
            if accessible(i_grid, x, y, False)
        ]

        if not different_positions:
            print("Error: I cannot remove a block from the next position")
            print_grid(i_grid)
            print()
            print_grid(g_grid)
            exit()

        (x, y) = max(
            different_positions,
            key=lambda w: (i_grid[w[1]][w[0]], -order_grid[w[1]][w[0]]),
        )
        # print("Current grid")
        # print_grid(i_grid)
        # print("Goal grid")
        # print_grid(g_grid)
        # print("Candidates: ", different_positions)
        # print ("Remove", (x, y))

        path = get_shortest_path(i_grid, depots[0], (x, y), False, order_grid)
        if not path:
            return None
        prev_loc = path[0]
        for loc in path[1:-1]:
            plan.append(move(i_grid, prev_loc, loc))
            prev_loc = loc

        plan.append(remove_block(i_grid, path[-2], path[-1]))

        for loc in path[::-1][2:]:
            plan.append(move(i_grid, prev_loc, loc))
            prev_loc = loc

        i_grid[y][x] -= 1

        plan.append(f"(destroy-block {depot_loc})")

    while plan[-1].startswith("(move"):
        plan.pop()

    return plan


def parse_grid_line(line, robots, depots, y):
    content = (
        line.replace("  ", " ").replace("  ", " ").replace(" \n", "")[1:].split(" ")
    )

    for (x, t) in enumerate(content):
        if "R" in t:
            robots.append((x, y))

        if "D" in t:
            depots.append((x, y))

    return map(lambda x: int(x.replace("R", "").replace("D", "")), content)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("raw_instance", type=argparse.FileType("r"))
    args = parser.parse_args()

    i_grid = []
    g_grid = []
    robots = []
    depots = []

    lines = args.raw_instance.readlines()
    instance_name = lines[0][:-1]

    i = 0
    while not lines[i].startswith("Initial state:"):
        i += 1

    i += 1
    y = 0
    while not lines[i].startswith("Goal state:"):
        i_grid.append(parse_grid_line(lines[i], robots, depots, y))
        y += 1
        i += 1

    i += 1

    while not lines[i].startswith("Maximal height:"):
        g_grid.append(parse_grid_line(lines[i], [], [], 0))
        i += 1

    grid_size_y = len(i_grid)
    grid_size_x = len(i_grid[0])
    max_height = int(lines[i].split(":")[1].strip())

    assert len(robots) == 1
    assert len(depots) == 1
    assert robots[0] == depots[0]

    plan = solve(grid_size_x, grid_size_y, max_height, robots, depots, i_grid, g_grid)
    print("\n".join(plan))


if __name__ == "__main__":
    main()
